# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import configparser
import logging
import os

from snapcraft.internal import errors


logger = logging.getLogger(__name__)


class DesktopFile:
    def __init__(
        self, *, name: str, filename: str, snap_name: str, prime_dir: str
    ) -> None:
        self._name = name
        self._filename = filename
        self._snap_name = snap_name
        self._prime_dir = prime_dir
        self._path = os.path.join(prime_dir, filename)
        if not os.path.exists(self._path):
            raise errors.InvalidDesktopFileError(
                filename, "does not exist (defined in the app {!r})".format(name)
            )

    def parse_and_reformat(self):
        self._parser = configparser.ConfigParser(interpolation=None)
        self._parser.optionxform = str
        self._parser.read(self._path, encoding="utf-8")
        section = "Desktop Entry"
        if section not in self._parser.sections():
            raise errors.InvalidDesktopFileError(
                self._filename, "missing 'Desktop Entry' section"
            )
        if "Exec" not in self._parser[section]:
            raise errors.InvalidDesktopFileError(self._filename, "missing 'Exec' key")
        # XXX: do we want to allow more parameters for Exec?
        if self._name == self._snap_name:
            exec_value = "{} %U".format(self._name)
        else:
            exec_value = "{}.{} %U".format(self._snap_name, self._name)
        self._parser[section]["Exec"] = exec_value
        if "Icon" in self._parser[section]:
            icon = self._parser[section]["Icon"]
            if icon.startswith("/"):
                icon = icon.lstrip("/")
                if os.path.exists(os.path.join(self._prime_dir, icon)):
                    self._parser[section]["Icon"] = "${{SNAP}}/{}".format(icon)
                else:
                    logger.warning(
                        "Icon {} specified in desktop file {} not found "
                        "in prime directory".format(icon, self._filename)
                    )

    def write(self, *, gui_dir: str) -> None:
        # Rename the desktop file to match the app name. This will help
        # unity8 associate them (https://launchpad.net/bugs/1659330).
        target_filename = "{}.desktop".format(self._name)
        target = os.path.join(gui_dir, target_filename)
        if os.path.exists(target):
            # Unlikely. A desktop file in setup/gui/ already existed for
            # this app. Let's pretend it wasn't there and overwrite it.
            os.remove(target)
        with open(target, "w", encoding="utf-8") as f:
            self._parser.write(f, space_around_delimiters=False)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from ._project_options import ProjectOptions
from ._project_info import ProjectInfo  # noqa: F401


class Project(ProjectOptions):
    """All details around building a project concerning the build environment
    and the snap being built."""

    def __init__(
        self,
        *,
        target_deb_arch=None,
        debug=False,
        snapcraft_yaml_file_path=None,
        work_dir: str = None,
        is_managed_host: bool = False
    ) -> None:

        project_dir = os.getcwd()
        if is_managed_host:
            work_dir = os.path.expanduser("~")
            internal_dir = work_dir
        else:
            work_dir = project_dir
            internal_dir = os.path.join(work_dir, "snap", ".snapcraft")

        # This here check is mostly for backwards compatibility with the
        # rest of the code base.
        if snapcraft_yaml_file_path is None:
            self.info = None  # type: ProjectInfo
        else:
            self.info = ProjectInfo(snapcraft_yaml_file_path=snapcraft_yaml_file_path)

        self._global_state_file = os.path.join(internal_dir, "state")
        self._project_dir = project_dir
        self._work_dir = work_dir
        self._internal_dir = internal_dir

        super().__init__(target_deb_arch, debug, work_dir=work_dir)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib

# This is used in a mypy 'type:' comment below, but nowhere else. Flake8
# doesn't like that very much, so noqa.
from typing import Dict  # noqa

from snapcraft.internal import errors

_ID_TO_UBUNTU_CODENAME = {
    "17.10": "artful",
    "17.04": "zesty",
    "16.04": "xenial",
    "14.04": "trusty",
}


class OsRelease:
    """A class to intelligently determine the OS on which we're running"""

    def __init__(self, *, os_release_file: str = "/etc/os-release") -> None:
        """Create a new OsRelease instance.

        :param str os_release_file: Path to os-release file to be parsed.
        """
        with contextlib.suppress(FileNotFoundError):
            self._os_release = {}  # type: Dict[str, str]
            with open(os_release_file) as f:
                for line in f:
                    entry = line.rstrip().split("=")
                    if len(entry) == 2:
                        self._os_release[entry[0]] = entry[1].strip('"')

    def id(self) -> str:
        """Return the OS ID

        :raises OsReleaseIdError: If no ID can be determined.
        """
        with contextlib.suppress(KeyError):
            return self._os_release["ID"]

        raise errors.OsReleaseIdError()

    def name(self) -> str:
        """Return the OS name

        :raises OsReleaseNameError: If no name can be determined.
        """
        with contextlib.suppress(KeyError):
            return self._os_release["NAME"]

        raise errors.OsReleaseNameError()

    def version_id(self) -> str:
        """Return the OS version ID

        :raises OsReleaseVersionIdError: If no version ID can be determined.
        """
        with contextlib.suppress(KeyError):
            return self._os_release["VERSION_ID"]

        raise errors.OsReleaseVersionIdError()

    def version_codename(self) -> str:
        """Return the OS version codename

        This first tries to use the VERSION_CODENAME. If that's missing, it
        tries to use the VERSION_ID to figure out the codename on its own.

        :raises OsReleaseCodenameError: If no version codename can be
                                        determined.
        """
        with contextlib.suppress(KeyError):
            return self._os_release["VERSION_CODENAME"]

        with contextlib.suppress(KeyError):
            return _ID_TO_UBUNTU_CODENAME[self._os_release["VERSION_ID"]]

        raise errors.OsReleaseCodenameError()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The scons plugin is useful for building parts that build with scons.

These are projects that have a SConstruct that drives the build.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - scons-options:
      (list of strings)
      flags to pass to the build using the scons semantics for parameters.
"""

import os

import snapcraft
from snapcraft.internal import errors


class SconsPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["scons-options"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]
        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["scons-options"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self._setup_base_tools(project.info.base)

    def _setup_base_tools(self, base):
        if base in ("core16", "core18"):
            self.build_packages.append("scons")
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def build(self):
        super().build()
        env = os.environ.copy()
        env["DESTDIR"] = self.installdir
        self.run(["scons"] + self.options.scons_options)
        self.run(["scons", "install"] + self.options.scons_options, env=env)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os.path


def setup_dirs() -> None:
    """
    Ensure that snapcraft.common plugindir is setup correctly
    and support running out of a development snapshot
    """
    from snapcraft.internal import common

    topdir = os.path.abspath(os.path.join(__file__, "..", "..", ".."))
    # Only change the default if we are running from a checkout or from the
    # snap.
    if os.path.exists(os.path.join(topdir, "setup.py")):
        common.set_plugindir(os.path.join(topdir, "snapcraft", "plugins"))
        common.set_schemadir(os.path.join(topdir, "schema"))
        common.set_extensionsdir(os.path.join(topdir, "extensions"))

    # The default paths are relative to sys.prefix, which works well for
    # Snapcraft as a deb or in a venv. However, the Python plugin installs
    # packages into $SNAP/ as a prefix, while Python itself is contained in
    # $SNAP/usr/. As a result, using sys.prefix (which is '/usr') to find these
    # files won't work in the snap.
    elif common.is_snap():
        parent_dir = os.path.join(os.environ.get("SNAP"), "share", "snapcraft")
        common.set_plugindir(os.path.join(parent_dir, "plugins"))
        common.set_schemadir(os.path.join(parent_dir, "schema"))
        common.set_extensionsdir(os.path.join(parent_dir, "extensions"))
        common.set_legacy_snapcraft_dir(
            os.path.join(os.environ.get("SNAP"), "legacy_snapcraft")
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import typing
from time import sleep

import click

from . import echo
from . import env
from ._options import add_build_options, get_project
from snapcraft.internal import (
    errors,
    build_providers,
    deprecations,
    lifecycle,
    project_loader,
    steps,
    repo,
)
from snapcraft.project._sanity_checks import (
    conduct_project_sanity_check,
    conduct_build_environment_sanity_check,
)
from snapcraft.project.errors import YamlValidationError, MultipassMissingLinuxError

if typing.TYPE_CHECKING:
    from snapcraft.internal.project import Project  # noqa: F401


def _install_multipass():
    repo.snaps.install_snaps(["multipass/latest/beta"])
    # wait for multipassd to be available
    click.echo("Waiting for multipass...")
    retry_count = 20
    while retry_count:
        try:
            output = subprocess.check_output(["multipass", "version"]).decode()
        except subprocess.CalledProcessError:
            output = ""
        # if multipassd is in the version information, it means the service is up
        # and we can carry on
        if "multipassd" in output:
            break
        retry_count -= 1
        sleep(1)
    # No need to worry about getting to this point by exhausting our retry count,
    # the rest of the stack will handle the error appropriately.


# TODO: when snap is a real step we can simplify the arguments here.
# fmt: off
def _execute(  # noqa: C901
    step: steps.Step,
    parts: str,
    pack_project: bool = False,
    output: str = None,
    shell: bool = False,
    shell_after: bool = False,
    destructive_mode: bool = False,
    **kwargs
) -> "Project":
    # fmt: on
    provider = "host" if destructive_mode else None
    build_environment = env.BuilderEnvironmentConfig(force_provider=provider)
    try:
        conduct_build_environment_sanity_check(build_environment.provider)
    except MultipassMissingLinuxError as e:
        if click.confirm(str(e)):
            _install_multipass()
        else:
            raise errors.SnapcraftEnvironmentError("multipass is required to continue.") from e

    project = get_project(is_managed_host=build_environment.is_managed_host, **kwargs)

    conduct_project_sanity_check(project)

    if build_environment.is_managed_host or build_environment.is_host:
        project_config = project_loader.load_config(project)
        lifecycle.execute(step, project_config, parts)
        if pack_project:
            _pack(project.prime_dir, output=output)
    else:
        build_provider_class = build_providers.get_provider_for(
            build_environment.provider
        )
        echo.info("Launching a VM.")
        with build_provider_class(project=project, echoer=echo) as instance:
            instance.mount_project()
            try:
                if shell:
                    # shell means we want to do everything right up to the previous
                    # step and then go into a shell instead of the requested step.
                    # the "snap" target is a special snowflake that has not made its
                    # way to be a proper step.
                    previous_step = None
                    if pack_project:
                        previous_step = steps.PRIME
                    elif step > steps.PULL:
                        previous_step = step.previous_step()
                    # steps.PULL is the first step, so we would directly shell into it.
                    if previous_step:
                        instance.execute_step(previous_step)
                elif pack_project:
                    instance.pack_project(output=output)
                else:
                    instance.execute_step(step)
            except Exception:
                if project.debug:
                    instance.shell()
                else:
                    echo.warning("Run the same command again with --debug to shell into the environment "
                                 "if you wish to introspect this failure.")
                    raise
            else:
                if shell or shell_after:
                    instance.shell()
    return project


def _pack(directory: str, *, output: str) -> None:
    snap_name = lifecycle.pack(directory, output)
    echo.info("Snapped {}".format(snap_name))


@click.group()
@add_build_options()
@click.pass_context
def lifecyclecli(ctx, **kwargs):
    pass


@lifecyclecli.command()
def init():
    """Initialize a snapcraft project."""
    snapcraft_yaml_path = lifecycle.init()
    echo.info("Created {}.".format(snapcraft_yaml_path))
    echo.wrapped("Go to https://docs.snapcraft.io/the-snapcraft-format/8337 for more "
                 "information about the snapcraft.yaml format.")


@lifecyclecli.command()
@click.pass_context
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def pull(ctx, parts, **kwargs):
    """Download or retrieve artifacts defined for a part.

    \b
    Examples:
        snapcraft pull
        snapcraft pull my-part1 my-part2

    """
    _execute(steps.PULL, parts, **kwargs)


@lifecyclecli.command()
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def build(parts, **kwargs):
    """Build artifacts defined for a part.

    \b
    Examples:
        snapcraft build
        snapcraft build my-part1 my-part2

    """
    _execute(steps.BUILD, parts, **kwargs)


@lifecyclecli.command()
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def stage(parts, **kwargs):
    """Stage the part's built artifacts into the common staging area.

    \b
    Examples:
        snapcraft stage
        snapcraft stage my-part1 my-part2

    """
    _execute(steps.STAGE, parts, **kwargs)


@lifecyclecli.command()
@add_build_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def prime(parts, **kwargs):
    """Final copy and preparation for the snap.

    \b
    Examples:
        snapcraft prime
        snapcraft prime my-part1 my-part2

    """
    _execute(steps.PRIME, parts, **kwargs)


@lifecyclecli.command()
@add_build_options()
@click.argument("directory", required=False)
@click.option("--output", "-o", help="path to the resulting snap.")
def snap(directory, output, **kwargs):
    """Create a snap.

    \b
    Examples:
        snapcraft snap
        snapcraft snap --output renamed-snap.snap

    If you want to snap a directory, you should use the pack command
    instead.
    """
    if directory:
        deprecations.handle_deprecation_notice("dn6")
        _pack(directory, output=output)
    else:
        _execute(steps.PRIME, parts=[], pack_project=True, output=output, **kwargs)


@lifecyclecli.command()
@click.argument("directory")
@click.option("--output", "-o", help="path to the resulting snap.")
def pack(directory, output, **kwargs):
    """Create a snap from a directory holding a valid snap.

    The layout of <directory> should contain a valid meta/snap.yaml in
    order to be a valid snap.

    \b
    Examples:
        snapcraft pack my-snap-directory
        snapcraft pack my-snap-directory --output renamed-snap.snap

    """
    _pack(directory, output=output)


@lifecyclecli.command()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
@click.option(
    "--step",
    "-s",
    "step_name",
    type=click.Choice(["pull", "build", "stage", "prime", "strip"]),
    help="only clean the specified step and those that depend on it.",
)
def clean(parts, step_name):
    """Remove content - cleans downloads, builds or install artifacts.

    \b
    Examples:
        snapcraft clean
        snapcraft clean my-part --step build
    """
    build_environment = env.BuilderEnvironmentConfig()
    try:
        project = get_project(
            is_managed_host=build_environment.is_managed_host
        )
    except YamlValidationError:
        # We need to be able to clean invalid projects too.
        project = get_project(
            is_managed_host=build_environment.is_managed_host,
            skip_snapcraft_yaml=True
        )

    step = None
    if step_name:
        if step_name == "strip":
            echo.warning(
                "DEPRECATED: Use `prime` instead of `strip` as the step to clean"
            )
            step_name = "prime"
        step = steps.get_step_by_name(step_name)

    if build_environment.is_host:
        lifecycle.clean(project, parts, step)
    else:
        # TODO support for steps.
        if parts or step_name:
            raise errors.SnapcraftEnvironmentError(
                "Build providers are still not feature complete, specifying parts or a step name "
                "is not yet supported.")
        build_provider_class = build_providers.get_provider_for(
            build_environment.provider
        )
        build_provider_class(project=project, echoer=echo).clean_project()


if __name__ == "__main__":
    lifecyclecli.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The cmake plugin is useful for building cmake based parts.

These are projects that have a CMakeLists.txt that drives the build.
The plugin requires a CMakeLists.txt in the root of the source tree.

This plugin also supports options from the `make` plugin. Run
`snapcraft help make` for more details.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - configflags:
      (list of strings)
      configure flags to pass to the build using the common cmake semantics.
"""

import os

import snapcraft.plugins.make


class CMakePlugin(snapcraft.plugins.make.MakePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["configflags"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return super().get_build_properties() + ["configflags"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)
        self.build_packages.append("cmake")
        self.out_of_source_build = True

    def build(self):
        source_subdir = getattr(self.options, "source_subdir", None)
        if source_subdir:
            sourcedir = os.path.join(self.sourcedir, source_subdir)
        else:
            sourcedir = self.sourcedir

        env = self._build_environment()

        self.run(
            ["cmake", sourcedir, "-DCMAKE_INSTALL_PREFIX="] + self.options.configflags,
            env=env,
        )

        self.make(env=env)

    def _build_environment(self):
        env = os.environ.copy()
        env["CMAKE_PREFIX_PATH"] = "$CMAKE_PREFIX_PATH:{}".format(
            self.project.stage_dir
        )
        env["CMAKE_INCLUDE_PATH"] = "$CMAKE_INCLUDE_PATH:" + ":".join(
            ["{0}/include", "{0}/usr/include", "{0}/include/{1}", "{0}/usr/include/{1}"]
        ).format(self.project.stage_dir, self.project.arch_triplet)
        env["CMAKE_LIBRARY_PATH"] = "$CMAKE_LIBRARY_PATH:" + ":".join(
            ["{0}/lib", "{0}/usr/lib", "{0}/lib/{1}", "{0}/usr/lib/{1}"]
        ).format(self.project.stage_dir, self.project.arch_triplet)

        return env



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from snapcraft import formatting_utils, project
from snapcraft.internal import common, elf, pluginhandler

from typing import Dict, List


def env_for_classic(base: str, arch_triplet: str) -> List[str]:
    """Set the required environment variables for a classic confined build."""
    env = []

    core_path = common.get_core_path(base)
    paths = common.get_library_paths(core_path, arch_triplet, existing_only=False)
    env.append(
        formatting_utils.format_path_variable(
            "LD_LIBRARY_PATH", paths, prepend="", separator=":"
        )
    )

    return env


def runtime_env(root: str, arch_triplet: str) -> List[str]:
    """Set the environment variables required for running binaries."""
    env = []

    env.append(
        'PATH="'
        + ":".join(
            ["{0}/usr/sbin", "{0}/usr/bin", "{0}/sbin", "{0}/bin", "$PATH"]
        ).format(root)
        + '"'
    )

    # Add the default LD_LIBRARY_PATH
    paths = common.get_library_paths(root, arch_triplet)
    # Add more specific LD_LIBRARY_PATH from staged packages if necessary
    paths += elf.determine_ld_library_path(root)

    if paths:
        env.append(
            formatting_utils.format_path_variable(
                "LD_LIBRARY_PATH", paths, prepend="", separator=":"
            )
        )

    return env


def build_env(root: str, snap_name: str, arch_triplet: str) -> List[str]:
    """Set the environment variables required for building.

    This is required for the current parts installdir due to stage-packages
    and also to setup the stagedir.
    """
    env = []

    paths = common.get_include_paths(root, arch_triplet)
    if paths:
        for envvar in ["CPPFLAGS", "CFLAGS", "CXXFLAGS"]:
            env.append(
                formatting_utils.format_path_variable(
                    envvar, paths, prepend="-I", separator=" "
                )
            )

    paths = common.get_library_paths(root, arch_triplet)
    if paths:
        env.append(
            formatting_utils.format_path_variable(
                "LDFLAGS", paths, prepend="-L", separator=" "
            )
        )

    paths = common.get_pkg_config_paths(root, arch_triplet)
    if paths:
        env.append(
            formatting_utils.format_path_variable(
                "PKG_CONFIG_PATH", paths, prepend="", separator=":"
            )
        )

    return env


def build_env_for_stage(stagedir: str, snap_name: str, arch_triplet: str) -> List[str]:
    env = build_env(stagedir, snap_name, arch_triplet)
    env.append('PERL5LIB="{0}/usr/share/perl5/"'.format(stagedir))

    return env


def snapcraft_global_environment(project: project.Project) -> Dict[str, str]:
    if project.info.name:
        name = project.info.name
    else:
        name = ""

    if project.info.version:
        version = project.info.version
    else:
        version = ""

    if project.info.grade:
        grade = project.info.grade
    else:
        grade = ""

    return {
        "SNAPCRAFT_ARCH_TRIPLET": project.arch_triplet,
        "SNAPCRAFT_PARALLEL_BUILD_COUNT": str(project.parallel_build_count),
        "SNAPCRAFT_PROJECT_NAME": name,
        "SNAPCRAFT_PROJECT_VERSION": version,
        "SNAPCRAFT_PROJECT_GRADE": grade,
        "SNAPCRAFT_STAGE": project.stage_dir,
        "SNAPCRAFT_PRIME": project.prime_dir,
        "SNAPCRAFT_EXTENSIONS_DIR": common.get_extensionsdir(),
    }


def snapcraft_part_environment(part: pluginhandler.PluginHandler) -> Dict[str, str]:
    return {
        "SNAPCRAFT_PART_SRC": part.plugin.sourcedir,
        "SNAPCRAFT_PART_BUILD": part.plugin.builddir,
        "SNAPCRAFT_PART_INSTALL": part.plugin.installdir,
    }


def environment_to_replacements(environment: Dict[str, str]) -> Dict[str, str]:
    replacements = {}  # type: Dict[str, str]
    for variable, value in environment.items():
        # Support both $VAR and ${VAR} syntax
        replacements["${}".format(variable)] = value
        replacements["${{{}}}".format(variable)] = value

    return replacements



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import json
import logging
import os
import sys

import click

from snapcraft.internal import errors
from snapcraft.cli._errors import exception_handler
from snapcraft.internal import log


@click.group()
@click.option("--debug", "-d", is_flag=True, envvar="SNAPCRAFT_ENABLE_DEVELOPER_DEBUG")
def run(debug):
    """snapcraftctl is how snapcraft.yaml can communicate with snapcraft"""

    if debug:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO

    # Setup global exception handler (to be called for unhandled exceptions)
    sys.excepthook = functools.partial(exception_handler, debug=debug)

    # In an ideal world, this logger setup would be replaced
    log.configure(log_level=log_level)


@run.command()
def pull():
    """Run the 'pull' step of the calling part's lifecycle"""
    _call_function("pull")


@run.command()
def build():
    """Run the 'build' step of the calling part's lifecycle"""
    _call_function("build")


@run.command()
def stage():
    """Run the 'stage' step of the calling part's lifecycle"""
    _call_function("stage")


@run.command()
def prime():
    """Run the 'prime' step of the calling part's lifecycle"""
    _call_function("prime")


@run.command("set-version")
@click.argument("version")
def set_version(version):
    """Set the version of the snap"""
    _call_function("set-version", {"version": version})


@run.command("set-grade")
@click.argument("grade")
def set_grade(grade):
    """Set the grade of the snap"""
    _call_function("set-grade", {"grade": grade})


def _call_function(function_name, args=None):
    if not args:
        args = {}

    data = {"function": function_name, "args": args}

    # We could load the FIFOs in `run` and shove them in the context, but
    # that's too early to error out if these variables aren't defined. Doing it
    # here allows one to run e.g. `snapcraftctl build --help` without needing
    # these variables defined, which is a win for usability.
    try:
        call_fifo = os.environ["SNAPCRAFTCTL_CALL_FIFO"]
        feedback_fifo = os.environ["SNAPCRAFTCTL_FEEDBACK_FIFO"]
    except KeyError as e:
        raise errors.SnapcraftEnvironmentError(
            "{!s} environment variable must be defined. Note that this "
            "utility is only designed for use within a snapcraft.yaml".format(e)
        ) from e

    with open(call_fifo, "w") as f:
        f.write(json.dumps(data))
        f.flush()

    with open(feedback_fifo, "r") as f:
        feedback = f.readline().strip()

    # Any feedback is considered a fatal error to be printed
    if feedback:
        raise errors.SnapcraftctlError(feedback)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, Set

from snapcraft import project, BasePlugin
from snapcraft.internal.project_loader import grammar
from snapcraft.internal import repo

from ._package_transformer import package_transformer


class PartGrammarProcessor:
    """Process part properties that support grammar.

    Stage packages example:
    >>> from unittest import mock
    >>> import snapcraft
    >>> # Pretend that all packages are valid
    >>> repo = mock.Mock()
    >>> repo.is_valid.return_value = True
    >>> plugin = mock.Mock()
    >>> plugin.stage_packages = [{'try': ['foo']}]
    >>> processor = PartGrammarProcessor(
    ...    plugin=plugin,
    ...    properties={},
    ...    project=snapcraft.project.Project(),
    ...    repo=repo)
    >>> processor.get_stage_packages()
    {'foo'}

    Build packages example:
    >>> from unittest import mock
    >>> import snapcraft
    >>> # Pretend that all packages are valid
    >>> repo = mock.Mock()
    >>> repo.is_valid.return_value = True
    >>> plugin = mock.Mock()
    >>> plugin.build_packages = [{'try': ['foo']}]
    >>> processor = PartGrammarProcessor(
    ...    plugin=plugin,
    ...    properties={},
    ...    project=snapcraft.project.Project(),
    ...    repo=repo)
    >>> processor.get_build_packages()
    {'foo'}

    Source example:
    >>> from unittest import mock
    >>> import snapcraft
    >>> plugin = mock.Mock()
    >>> plugin.properties = {'source': [{'on amd64': 'foo'}, 'else fail']}
    >>> processor = PartGrammarProcessor(
    ...    plugin=plugin,
    ...    properties=plugin.properties,
    ...    project=snapcraft.project.Project(),
    ...    repo=None)
    >>> processor.get_source()
    'foo'
    """

    def __init__(
        self,
        *,
        plugin: BasePlugin,
        properties: Dict[str, Any],
        project: project.Project,
        repo: "repo.Ubuntu"
    ) -> None:
        self._plugin = plugin
        self._project = project
        self._repo = repo

        self.__build_snaps = set()  # type: Set[str]
        self.__build_packages = set()  # type: Set[str]
        self.__stage_packages = set()  # type: Set[str]

        source_grammar = properties.get("source", [""])
        if not isinstance(source_grammar, list):
            self._source_grammar = [source_grammar]
        else:
            self._source_grammar = source_grammar

        self.__source = ""

    def get_source(self) -> str:
        if not self.__source:
            # The grammar is array-based, even though we only support a single
            # source.
            processor = grammar.GrammarProcessor(
                self._source_grammar, self._project, lambda s: True
            )
            source_array = processor.process()
            if len(source_array) > 0:
                self.__source = source_array.pop()
        return self.__source

    def get_build_snaps(self) -> Set[str]:
        if not self.__build_snaps:
            processor = grammar.GrammarProcessor(
                getattr(self._plugin, "build_snaps", []),
                self._project,
                repo.snaps.SnapPackage.is_valid_snap,
            )
            self.__build_snaps = processor.process()

        return self.__build_snaps

    def get_build_packages(self) -> Set[str]:
        if not self.__build_packages:
            processor = grammar.GrammarProcessor(
                getattr(self._plugin, "build_packages", []),
                self._project,
                self._repo.build_package_is_valid,
                transformer=package_transformer,
            )
            self.__build_packages = processor.process()

        return self.__build_packages

    def get_stage_packages(self) -> Set[str]:
        if not self.__stage_packages:
            processor = grammar.GrammarProcessor(
                getattr(self._plugin, "stage_packages", []),
                self._project,
                self._repo.is_valid,
                transformer=package_transformer,
            )
            self.__stage_packages = processor.process()

        return self.__stage_packages



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import click
from datetime import datetime
import json
import logging
from tabulate import tabulate

from snapcraft import formatting_utils
from snapcraft.internal import errors, project_loader
from snapcraft.internal.project_loader import inspection
from ._options import get_project
from . import echo


@click.group(context_settings={})
@click.pass_context
def inspectcli(ctx):
    pass


@inspectcli.command()
@click.option(
    "--provides",
    metavar="<path>",
    help=(
        "Show the part that provided <path>. <path> can be relative or absolute, and "
        "must be pointing within the staging or priming area."
    ),
)
@click.option(
    "--latest-step", is_flag=True, help="Show the most recently-completed step"
)
@click.option("--json", "use_json", is_flag=True, help="Print the result in json")
def inspect(*, provides: str, latest_step: bool, use_json: bool, **kwargs):
    """Inspect the current state of the project.

    By default, this command will print a tabulated summary of the current state of the
    project.
    """
    count = 0
    for option in (provides, latest_step):
        if option:
            count += 1

    if count > 1:
        raise errors.SnapcraftEnvironmentError(
            "Only one of --provides or --latest-step may be supplied"
        )

    # Silence the plugin loader logger
    logging.getLogger("snapcraft.internal.pluginhandler._plugin_loader").setLevel(
        logging.ERROR
    )
    project = get_project(**kwargs)
    config = project_loader.load_config(project)

    if provides:
        providing_parts = inspection.provides(
            provides, config.project, config.all_parts
        )
        providing_part_names = sorted([p.name for p in providing_parts])
        if use_json:
            print(
                json.dumps(
                    {"path": provides, "parts": providing_part_names},
                    sort_keys=True,
                    indent=4,
                )
            )
        else:
            echo.info(
                "This path was provided by the following {}:".format(
                    formatting_utils.pluralize(providing_part_names, "part", "parts")
                )
            )
            echo.info("\n".join(providing_part_names))

    elif latest_step:
        part, step, timestamp = inspection.latest_step(config.all_parts)
        directory = part.working_directory_for_step(step)
        if use_json:
            print(
                json.dumps(
                    {
                        "part": part.name,
                        "step": step.name,
                        "directory": directory,
                        "timestamp": timestamp,
                    },
                    sort_keys=True,
                    indent=4,
                )
            )
        else:
            echo.info(
                "The latest step that was run is the {!r} step of the {!r} part, "
                "which ran at {}. The working directory for this step is {!r}".format(
                    step.name,
                    part.name,
                    datetime.fromtimestamp(timestamp).strftime("%c"),
                    directory,
                )
            )

    else:
        summary = inspection.lifecycle_status(config)
        if use_json:
            summary_dict = dict()
            for part_summary in summary:
                part_name = part_summary.pop("part")
                summary_dict[part_name] = part_summary
            print(json.dumps(summary_dict, sort_keys=True, indent=4))
        else:
            print(tabulate(summary, headers="keys"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import glob
import os
from textwrap import dedent

from ._python_finder import get_python_command
from . import errors

_SITECUSTOMIZE_TEMPLATE = dedent(
    """\
    import site
    import os

    snap_dir = os.getenv("SNAP")
    snapcraft_stage_dir = os.getenv("SNAPCRAFT_STAGE")
    snapcraft_part_install = os.getenv("SNAPCRAFT_PART_INSTALL")

    for d in (snap_dir, snapcraft_stage_dir, snapcraft_part_install):
        if d:
            site_dir = os.path.join(d, "{site_dir}")
            site.addsitedir(site_dir)

    if snap_dir:
        site.ENABLE_USER_SITE = False"""
)


def _get_user_site_dir(python_major_version, *, install_dir):
    path_glob = os.path.join(
        install_dir, "lib", "python{}*".format(python_major_version), "site-packages"
    )
    user_site_dirs = glob.glob(path_glob)
    if not user_site_dirs:
        raise errors.MissingUserSitePackagesError(path_glob)

    return user_site_dirs[0][len(install_dir) + 1 :]


def _get_sitecustomize_path(python_major_version, *, stage_dir, install_dir):
    # Use the part's install_dir unless there's a python in the staging area
    base_dir = install_dir
    with contextlib.suppress(errors.MissingPythonCommandError):
        python_command = get_python_command(
            python_major_version, stage_dir=stage_dir, install_dir=install_dir
        )
        if python_command.startswith(stage_dir):
            base_dir = stage_dir

    site_py_glob = os.path.join(
        base_dir, "usr", "lib", "python{}*".format(python_major_version), "site.py"
    )
    python_sites = glob.glob(site_py_glob)
    if not python_sites:
        raise errors.MissingSitePyError(site_py_glob)

    python_site_dir = os.path.dirname(python_sites[0])

    return os.path.join(
        install_dir, python_site_dir[len(base_dir) + 1 :], "sitecustomize.py"
    )


def generate_sitecustomize(python_major_version, *, stage_dir, install_dir):
    """Generate a sitecustomize.py to look in staging, part install, and snap.

    This is done by checking the values of the environment variables $SNAP,
    $SNAPCRAFT_STAGE, and $SNAPCRAFT_PART_INSTALL. As a result, the same
    sitecustomize.py works to find packages at both build- and run-time.

    :param str python_major_version: The python major version to use (2 or 3)
    :param str stage_dir: Path to the staging area
    :param str install_dir: Path to the part's install area

    :raises MissingUserSitePackagesError: If no user site packages are found in
                                          install_dir/lib/pythonX*
    :raises MissingSitePyError: If no site.py can be found in either the
                                staging area or the part install area.
    """
    sitecustomize_path = _get_sitecustomize_path(
        python_major_version, stage_dir=stage_dir, install_dir=install_dir
    )
    os.makedirs(os.path.dirname(sitecustomize_path), exist_ok=True)

    # There may very well already be a sitecustomize.py already there. If so,
    # get rid of it. Is may be a symlink to another sitecustomize.py, in which
    # case, we'll get rid of that one as well.
    if os.path.islink(sitecustomize_path):
        target_path = os.path.realpath(sitecustomize_path)

        # Only remove the target if it's contained within the install directory
        if target_path.startswith(os.path.abspath(install_dir) + os.sep):
            with contextlib.suppress(FileNotFoundError):
                os.remove(target_path)

    with contextlib.suppress(FileNotFoundError):
        os.remove(sitecustomize_path)

    # Create our sitecustomize. Python from the archives already has one
    # which is distro-specific and not needed here, so we truncate it if it's
    # already there.
    with open(sitecustomize_path, "w") as f:
        f.write(
            _SITECUSTOMIZE_TEMPLATE.format(
                site_dir=_get_user_site_dir(
                    python_major_version, install_dir=install_dir
                )
            )
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
from collections import OrderedDict
from copy import deepcopy

import yaml

from snapcraft import yaml_utils
from . import _schema
from . import errors


class ProjectInfo:
    """Information gained from the snap's snapcraft.yaml file."""

    def __init__(self, *, snapcraft_yaml_file_path) -> None:
        self.snapcraft_yaml_file_path = snapcraft_yaml_file_path
        self.__raw_snapcraft = _load_yaml(yaml_file_path=snapcraft_yaml_file_path)

        try:
            self.name = self.__raw_snapcraft["name"]
        except KeyError as key_error:
            raise errors.YamlValidationError(
                "'name' is a required property in {!r}".format(snapcraft_yaml_file_path)
            ) from key_error
        self.version = self.__raw_snapcraft.get("version")
        self.summary = self.__raw_snapcraft.get("summary")
        self.description = self.__raw_snapcraft.get("description")
        self.confinement = self.__raw_snapcraft.get("confinement")
        self.architectures = self.__raw_snapcraft.get("architectures")
        self.grade = self.__raw_snapcraft.get("grade")
        self.base = self.__raw_snapcraft.get("base")
        self.type = self.__raw_snapcraft.get("type")

    def validate_raw_snapcraft(self):
        """Validate the snapcraft.yaml for this project."""
        _schema.Validator(self.__raw_snapcraft).validate()

    def get_raw_snapcraft(self):
        # TODO this should be a MappingProxyType, but ordered writing
        #      depends on reading in the current code base.
        return deepcopy(self.__raw_snapcraft)


def _load_yaml(*, yaml_file_path: str) -> OrderedDict:
    with open(yaml_file_path, "rb") as fp:
        bs = fp.read(2)

    if bs == codecs.BOM_UTF16_LE or bs == codecs.BOM_UTF16_BE:
        encoding = "utf-16"
    else:
        encoding = "utf-8"

    try:
        with open(yaml_file_path, encoding=encoding) as fp:  # type: ignore
            yaml_contents = yaml_utils.load(fp)  # type: ignore
    except yaml.MarkedYAMLError as e:
        raise errors.YamlValidationError(
            "{} on line {}, column {}".format(
                e.problem, e.problem_mark.line + 1, e.problem_mark.column + 1
            ),
            yaml_file_path,
        ) from e
    except yaml.reader.ReaderError as e:
        raise errors.YamlValidationError(
            "invalid character {!r} at position {}: {}".format(
                chr(e.character), e.position + 1, e.reason
            ),
            yaml_file_path,
        ) from e
    except yaml.YAMLError as e:
        raise errors.YamlValidationError(str(e), yaml_file_path) from e

    return yaml_contents



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import subprocess

from snapcraft import shell_utils
from snapcraft.internal.meta import _errors
from snapcraft.internal import sources


logger = logging.getLogger(__name__)


def get_version(version: str, version_script: str = None) -> str:
    new_version = version
    if version_script:
        logger.info("Determining the version from the project repo (version-script).")
        try:
            new_version = shell_utils.run_script(version_script).strip()
            if not new_version:
                raise _errors.CommandError("The version-script produced no output")
        except subprocess.CalledProcessError as e:
            raise _errors.CommandError(
                "The version-script failed to run (exit code {})".format(e.returncode)
            )
    # we want to whitelist what we support here.
    elif version == "git":
        logger.info("Determining the version from the project repo (version: git).")
        vcs_handler = sources.get_source_handler_from_type("git")
        new_version = vcs_handler.generate_version()

    if new_version != version:
        logger.info("The version has been set to {!r}".format(new_version))
    return new_version



#!/usr/bin/python3
# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import jsonschema

from . import errors
from snapcraft import yaml_utils
from snapcraft.internal import common


class Validator:
    def __init__(self, snapcraft_yaml=None):
        """Create a validation instance for snapcraft_yaml."""
        self._snapcraft = snapcraft_yaml if snapcraft_yaml else {}
        self._load_schema()

    @property
    def schema(self):
        """Return all schema properties."""

        return self._schema["properties"].copy()

    @property
    def part_schema(self):
        """Return part-specific schema properties."""

        sub = self.schema["parts"]["patternProperties"]
        properties = sub["^(?!plugins$)[a-z0-9][a-z0-9+-]*$"]["properties"]
        return properties

    @property
    def definitions_schema(self):
        """Return sub-schema that describes definitions used within schema."""

        return self._schema["definitions"].copy()

    def _load_schema(self):
        schema_file = os.path.abspath(
            os.path.join(common.get_schemadir(), "snapcraft.yaml")
        )
        try:
            with open(schema_file) as fp:
                self._schema = yaml_utils.load(fp)
        except FileNotFoundError:
            raise errors.YamlValidationError(
                "snapcraft validation file is missing from installation path"
            )

    def validate(self, *, source="snapcraft.yaml"):
        format_check = jsonschema.FormatChecker()
        try:
            jsonschema.validate(
                self._snapcraft, self._schema, format_checker=format_check
            )
        except jsonschema.ValidationError as e:
            raise errors.YamlValidationError.from_validation_error(e, source=source)



import json
import os

import requests
from simplejson.scanner import JSONDecodeError

from ._client import Client

from . import errors
from . import constants


class SSOClient(Client):
    """The Single Sign On server deals with authentication.
    It is used directly or indirectly by other servers.
    """

    def __init__(self, conf):
        super().__init__(
            conf,
            os.environ.get(
                "UBUNTU_SSO_API_ROOT_URL", constants.UBUNTU_SSO_API_ROOT_URL
            ),
        )

    def get_unbound_discharge(self, email, password, one_time_password, caveat_id):
        data = dict(email=email, password=password, caveat_id=caveat_id)
        if one_time_password:
            data["otp"] = one_time_password
        response = self.post(
            "tokens/discharge",
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        try:
            response_json = response.json()
        except JSONDecodeError:
            response_json = {}
        if response.ok:
            return response_json["discharge_macaroon"]
        else:
            if response.status_code == requests.codes.unauthorized and any(
                error.get("code") == "twofactor-required"
                for error in response_json.get("error_list", [])
            ):
                raise errors.StoreTwoFactorAuthenticationRequired()
            else:
                raise errors.StoreAuthenticationError(
                    "Failed to get unbound discharge", response
                )

    def refresh_unbound_discharge(self, unbound_discharge):
        data = {"discharge_macaroon": unbound_discharge}
        response = self.post(
            "tokens/refresh",
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if response.ok:
            return response.json()["discharge_macaroon"]
        else:
            raise errors.StoreAuthenticationError(
                "Failed to refresh unbound discharge", response
            )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import re
import subprocess
from typing import FrozenSet

from snapcraft import file_utils
from snapcraft.internal import elf


logger = logging.getLogger(__name__)


def rewrite_python_shebangs(root_dir):
    """Recursively change #!/usr/bin/pythonX shebangs to #!/usr/bin/env pythonX

    :param str root_dir: Directory that will be crawled for shebangs.
    """

    file_pattern = re.compile(r"")
    argless_shebang_pattern = re.compile(r"\A#!.*(python\S*)$", re.MULTILINE)
    shebang_pattern_with_args = re.compile(
        r"\A#!.*(python\S*)[ \t\f\v]+(\S+)$", re.MULTILINE
    )

    file_utils.replace_in_file(
        root_dir, file_pattern, argless_shebang_pattern, r"#!/usr/bin/env \1"
    )

    # The above rewrite will barf if the shebang includes any args to python.
    # For example, if the shebang was `#!/usr/bin/python3 -Es`, just replacing
    # that with `#!/usr/bin/env python3 -Es` isn't going to work as `env`
    # doesn't support arguments like that.
    #
    # The solution is to replace the shebang with one pointing to /bin/sh, and
    # then exec the original shebang with included arguments. This requires
    # some quoting hacks to ensure the file can be interpreted by both sh as
    # well as python, but it's better than shipping our own `env`.
    file_utils.replace_in_file(
        root_dir,
        file_pattern,
        shebang_pattern_with_args,
        r"""#!/bin/sh\n''''exec \1 \2 -- "$0" "$@" # '''""",
    )


def clear_execstack(*, elf_files: FrozenSet[elf.ElfFile]) -> None:
    """Clears the execstack for the relevant elf_files

    param elf.ElfFile elf_files: the full list of elf files to analyze
                                 and clear the execstack if present.
    """
    execstack_path = file_utils.get_tool_path("execstack")
    elf_files_with_execstack = [e for e in elf_files if e.execstack_set]

    if elf_files_with_execstack:
        formatted_items = ["- {}".format(e.path) for e in elf_files_with_execstack]
        logger.warning(
            "The execstacks are going to be cleared for the following "
            "files:\n{}\n"
            "To disable this behavior set "
            "`build-attributes: [keep-execstack]` "
            "for the part.".format("\n".join(formatted_items))
        )

    for elf_file in elf_files_with_execstack:
        try:
            subprocess.check_call([execstack_path, "--clear-execstack", elf_file.path])
        except subprocess.CalledProcessError:
            logger.warning("Failed to clear execstack for {!r}".format(elf_file.path))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager, suppress
import errno
import hashlib
import logging
import re
import os
import shutil
import subprocess
import sys
from typing import Pattern, Callable, Generator, List
from typing import Set  # noqa F401

from snapcraft.internal import common
from snapcraft.internal.errors import (
    RequiredCommandFailure,
    RequiredCommandNotFound,
    RequiredPathDoesNotExist,
    SnapcraftEnvironmentError,
    SnapcraftCopyFileNotFoundError,
    ToolMissingError,
)

if sys.version_info < (3, 6):
    import sha3  # noqa


logger = logging.getLogger(__name__)


def replace_in_file(
    directory: str, file_pattern: Pattern, search_pattern: Pattern, replacement: str
) -> None:
    """Searches and replaces patterns that match a file pattern.

    :param str directory: The directory to look for files.
    :param str file_pattern: The file pattern to match inside directory.
    :param search_pattern: A re.compile'd pattern to search for within
                           matching files.
    :param str replacement: The string to replace the matching search_pattern
                            with.
    """

    for root, directories, files in os.walk(directory):
        for file_name in files:
            if file_pattern.match(file_name):
                file_path = os.path.join(root, file_name)
                # Don't bother trying to rewrite a symlink. It's either invalid
                # or the linked file will be rewritten on its own.
                if not os.path.islink(file_path):
                    search_and_replace_contents(file_path, search_pattern, replacement)


def search_and_replace_contents(
    file_path: str, search_pattern: Pattern, replacement: str
) -> None:
    """Search file and replace any occurrence of pattern with replacement.

    :param str file_path: Path of file to be searched.
    :param re.RegexObject search_pattern: Pattern for which to search.
    :param str replacement: The string to replace pattern.
    """
    try:
        with open(file_path, "r+") as f:
            try:
                original = f.read()
            except UnicodeDecodeError:
                # This was probably a binary file. Skip it.
                return

            replaced = search_pattern.sub(replacement, original)
            if replaced != original:
                f.seek(0)
                f.truncate()
                f.write(replaced)
    except PermissionError as e:
        logger.warning(
            "Unable to open {path} for writing: {error}".format(path=file_path, error=e)
        )


def link_or_copy(source: str, destination: str, follow_symlinks: bool = False) -> None:
    """Hard-link source and destination files. Copy if it fails to link.

    Hard-linking may fail (e.g. a cross-device link, or permission denied), so
    as a backup plan we just copy it.

    :param str source: The source to which destination will be linked.
    :param str destination: The destination to be linked to source.
    :param bool follow_symlinks: Whether or not symlinks should be followed.
    """

    try:
        link(source, destination, follow_symlinks=follow_symlinks)
    except OSError as e:
        if e.errno == errno.EEXIST and not os.path.isdir(destination):
            # os.link will fail if the destination already exists, so let's
            # remove it and try again.
            os.remove(destination)
            link_or_copy(source, destination, follow_symlinks)
        else:
            copy(source, destination, follow_symlinks=follow_symlinks)


def link(source: str, destination: str, *, follow_symlinks: bool = False) -> None:
    """Hard-link source and destination files.

    :param str source: The source to which destination will be linked.
    :param str destination: The destination to be linked to source.
    :param bool follow_symlinks: Whether or not symlinks should be followed.

    :raises SnapcraftCopyFileNotFoundError: If source doesn't exist.
    """
    # Note that follow_symlinks doesn't seem to work for os.link, so we'll
    # implement this logic ourselves using realpath.
    source_path = source
    if follow_symlinks:
        source_path = os.path.realpath(source)

    if not os.path.exists(os.path.dirname(destination)):
        create_similar_directory(
            os.path.dirname(source_path), os.path.dirname(destination)
        )
    # Setting follow_symlinks=False in case this bug is ever fixed
    # upstream-- we want this function to continue supporting NOT following
    # symlinks.
    try:
        os.link(source_path, destination, follow_symlinks=False)
    except FileNotFoundError:
        raise SnapcraftCopyFileNotFoundError(source)


def copy(source: str, destination: str, *, follow_symlinks: bool = False) -> None:
    """Copy source and destination files.

    This function overwrites the destination if it already exists, and also
    tries to copy ownership information.

    :param str source: The source to be copied to destination.
    :param str destination: Where to put the copy.
    :param bool follow_symlinks: Whether or not symlinks should be followed.

    :raises SnapcraftCopyFileNotFoundError: If source doesn't exist.
    """
    # If os.link raised an I/O error, it may have left a file behind. Skip on
    # OSError in case it doesn't exist or is a directory.
    with suppress(OSError):
        os.unlink(destination)

    try:
        shutil.copy2(source, destination, follow_symlinks=follow_symlinks)
    except FileNotFoundError:
        raise SnapcraftCopyFileNotFoundError(source)
    uid = os.stat(source, follow_symlinks=follow_symlinks).st_uid
    gid = os.stat(source, follow_symlinks=follow_symlinks).st_gid
    try:
        os.chown(destination, uid, gid, follow_symlinks=follow_symlinks)
    except PermissionError as e:
        logger.debug(
            "Unable to chown {destination}: {error}".format(
                destination=destination, error=e
            )
        )


def link_or_copy_tree(
    source_tree: str,
    destination_tree: str,
    ignore: Callable[[str, List[str]], List[str]] = None,
    copy_function: Callable[..., None] = link_or_copy,
) -> None:
    """Copy a source tree into a destination, hard-linking if possible.

    :param str source_tree: Source directory to be copied.
    :param str destination_tree: Destination directory. If this directory
                                 already exists, the files in `source_tree`
                                 will take precedence.
    :param callable ignore: If given, called with two params, source dir and
                            dir contents, for every dir copied. Should return
                            list of contents to NOT copy.
    :param callable copy_function: Callable that actually copies.
    """

    if not os.path.isdir(source_tree):
        raise NotADirectoryError("{!r} is not a directory".format(source_tree))

    if not os.path.isdir(destination_tree) and (
        os.path.exists(destination_tree) or os.path.islink(destination_tree)
    ):
        raise NotADirectoryError(
            "Cannot overwrite non-directory {!r} with directory "
            "{!r}".format(destination_tree, source_tree)
        )

    create_similar_directory(source_tree, destination_tree)

    destination_basename = os.path.basename(destination_tree)

    for root, directories, files in os.walk(source_tree, topdown=True):
        ignored = set()  # type: Set[str]
        if ignore is not None:
            ignored = set(ignore(root, directories + files))

        # Don't recurse into destination tree if it's a subdirectory of the
        # source tree.
        if os.path.relpath(destination_tree, root) == destination_basename:
            ignored.add(destination_basename)

        if ignored:
            # Prune our search appropriately given an ignore list, i.e. don't
            # walk into directories that are ignored.
            directories[:] = [d for d in directories if d not in ignored]

        for directory in directories:
            source = os.path.join(root, directory)
            # os.walk doesn't by default follow symlinks (which is good), but
            # it includes symlinks that are pointing to directories in the
            # directories list. We want to treat it as a file, here.
            if os.path.islink(source):
                files.append(directory)
                continue

            destination = os.path.join(
                destination_tree, os.path.relpath(source, source_tree)
            )

            create_similar_directory(source, destination)

        for file_name in set(files) - ignored:
            source = os.path.join(root, file_name)
            destination = os.path.join(
                destination_tree, os.path.relpath(source, source_tree)
            )

            copy_function(source, destination)


def create_similar_directory(
    source: str, destination: str, follow_symlinks: bool = False
) -> None:
    """Create a directory with the same permission bits and owner information.

    :param str source: Directory from which to copy name, permission bits, and
                       owner information.
    :param str destintion: Directory to create and to which the `source`
                           information will be copied.
    :param bool follow_symlinks: Whether or not symlinks should be followed.
    """

    stat = os.stat(source, follow_symlinks=follow_symlinks)
    uid = stat.st_uid
    gid = stat.st_gid
    os.makedirs(destination, exist_ok=True)
    try:
        os.chown(destination, uid, gid, follow_symlinks=follow_symlinks)
    except PermissionError as exception:
        logger.debug("Unable to chown {}: {}".format(destination, exception))

    shutil.copystat(source, destination, follow_symlinks=follow_symlinks)


def executable_exists(path: str) -> bool:
    """Return True if 'path' exists and is readable and executable."""
    return os.path.exists(path) and os.access(path, os.R_OK | os.X_OK)


@contextmanager
def requires_command_success(
    command: str, not_found_fmt: str = None, failure_fmt: str = None
) -> Generator:
    if isinstance(command, str):
        cmd_list = command.split()
    else:
        raise TypeError("command must be a string.")
    kwargs = dict(command=command, cmd_list=cmd_list)
    try:
        subprocess.check_call(cmd_list, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except FileNotFoundError:
        if not_found_fmt is not None:
            kwargs["fmt"] = not_found_fmt
        raise RequiredCommandNotFound(**kwargs)
    except subprocess.CalledProcessError:
        if failure_fmt is not None:
            kwargs["fmt"] = failure_fmt
        raise RequiredCommandFailure(**kwargs)
    yield


@contextmanager
def requires_path_exists(path: str, error_fmt: str = None) -> Generator:
    if not os.path.exists(path):
        kwargs = dict(path=path)
        if error_fmt is not None:
            kwargs["fmt"] = error_fmt
        raise RequiredPathDoesNotExist(**kwargs)
    yield


def calculate_sha3_384(path: str) -> str:
    """Calculate sha3 384 hash, reading the file in 1MB chunks."""
    return calculate_hash(path, algorithm="sha3_384")


def calculate_hash(path: str, *, algorithm: str) -> str:
    """Calculate the hash for path with algorithm."""
    # This will raise an AttributeError if algorithm is unsupported
    hasher = getattr(hashlib, algorithm)()

    blocksize = 2 ** 20
    with open(path, "rb") as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            hasher.update(buf)
    return hasher.hexdigest()


def get_tool_path(command_name: str) -> str:
    """Return the path to the given command

    Return a path to command_name, if Snapcraft is running out of the snap
    or in legacy mode (snap or sources), it ensures it is using the one in
    the snap, not the host.
    If a path cannot be resolved, ToolMissingError is raised.

    : param str command_name: the name of the command to resolve a path for.
    :raises ToolMissingError: if command_name cannot be resolved to a path.
    :return: Path to command
    :rtype: str
    """
    if common.is_snap():
        command_path = _command_path_in_root(os.getenv("SNAP"), command_name)
    else:
        command_path = shutil.which(command_name)

    # shutil.which will return None if it cannot find command_name but
    # _command_path_in_root will return an empty string.
    if not command_path:
        raise ToolMissingError(command_name=command_name)

    return command_path


def _command_path_in_root(root, command_name: str) -> str:
    for bin_directory in (
        os.path.join("usr", "local", "sbin"),
        os.path.join("usr", "local", "bin"),
        os.path.join("usr", "sbin"),
        os.path.join("usr", "bin"),
        os.path.join("sbin"),
        os.path.join("bin"),
    ):
        path = os.path.join(root, bin_directory, command_name)
        if os.path.exists(path):
            return path

    return ""


def get_linker_version_from_file(linker_file: str) -> str:
    """Returns the version of the linker from linker_file.

    linker_file must be of the format ld-(?P<linker_version>[\d.]+).so$

    :param str linker_file: a valid file path or basename representing
                            the linker from libc6 or related.
    :returns: the version extracted from the linker file.
    :rtype: string
    :raises snapcraft.internal.errors.SnapcraftEnvironmentError:
       if linker_file is not of the expected format.
    """
    m = re.search(r"ld-(?P<linker_version>[\d.]+).so$", linker_file)
    if not m:
        # This is a programmatic error, we don't want to be friendly
        # about this.
        raise SnapcraftEnvironmentError(
            "The format for the linker should be of the of the form "
            "<root>/ld-<X>.<Y>.so. {!r} does not match that format. "
            "Ensure you are targeting an appropriate base".format(linker_file)
        )
    linker_version = m.group("linker_version")

    return linker_version



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from ._metadata import ExtractedMetadata
from snapcraft.extractors import _errors

from xml.etree.ElementTree import ElementTree, ParseError


def extract(path: str) -> ExtractedMetadata:
    if not path.endswith(".metainfo.xml") and not path.endswith(".appdata.xml"):
        raise _errors.UnhandledFileError(path, "appstream")

    try:
        tree = ElementTree().parse(path)
    except ParseError as e:
        raise _errors.AppstreamFileParseError(path) from e

    common_id = _get_value_from_xml_element(tree, "id")
    summary = _get_value_from_xml_element(tree, "summary")
    description = _get_value_from_xml_element(tree, "description")

    icon = None
    node = tree.find("icon")
    if node is not None and "type" in node.attrib and node.attrib["type"] == "local":
        # TODO Currently we only support local icons.
        # See bug https://bugs.launchpad.net/snapcraft/+bug/1742348
        # for supporting remote icons.
        # --elopio -20180109
        icon = node.text.strip()

    desktop_file_paths = []
    nodes = tree.findall("launchable")
    for node in nodes:
        if "type" in node.attrib and node.attrib["type"] == "desktop-id":
            desktop_file_id = node.text.strip()
            desktop_file_path = _desktop_file_id_to_path(desktop_file_id)
            if desktop_file_path:
                desktop_file_paths.append(desktop_file_path)

    return ExtractedMetadata(
        common_id=common_id,
        summary=summary,
        description=description,
        icon=icon,
        desktop_file_paths=desktop_file_paths,
    )


def _get_value_from_xml_element(tree, key) -> str:
    node = tree.find(key)
    if node is not None:
        return node.text.strip()
    else:
        return None


def _desktop_file_id_to_path(desktop_file_id: str) -> str:
    # For details about desktop file ids and their corresponding paths, see
    # https://standards.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html#desktop-file-id  # noqa
    for xdg_data_dir in ("usr/local/share", "usr/share"):
        desktop_file_path = os.path.join(
            xdg_data_dir, "applications", desktop_file_id.replace("-", "/")
        )
        if os.path.exists(desktop_file_path):
            return desktop_file_path
    else:
        return None



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Iterable, List, Sized


def combine_paths(paths: List[str], prepend: str, separator: str) -> str:
    """Combine list of paths into a string.

    :param list paths: List of paths to stringify.
    :param str prepend: String to prepend to each path in the string.
    :param str separator: String to place between each path in the string.
    """

    paths = ["{}{}".format(prepend, p) for p in paths]
    return separator.join(paths)


def format_path_variable(
    envvar: str, paths: List[str], prepend: str, separator: str
) -> str:
    """Return a path-like environment variable definition that appends.

    :param str envvar: The environment variable in question.
    :param list paths: The paths to append to the environment variable.
    :param str prepend: String to prepend to each path in the definition.
    :param str separator: String to place between each path in the definition.
    """

    if not paths:
        raise ValueError("Failed to format '${}': no paths supplied".format(envvar))

    return '{envvar}="${envvar}{separator}{paths}"'.format(
        envvar=envvar,
        separator=separator,
        paths=combine_paths(paths, prepend, separator),
    )


def humanize_list(
    items: Iterable[str], conjunction: str, item_format: str = "{!r}"
) -> str:
    """Format a list into a human-readable string.

    :param list items: List to humanize.
    :param str conjunction: The conjunction used to join the final element to
                            the rest of the list (e.g. 'and').
    :param str item_format: Format string to use per item.
    """

    if not items:
        return ""

    quoted_items = [item_format.format(item) for item in sorted(items)]
    if len(quoted_items) == 1:
        return quoted_items[0]

    humanized = ", ".join(quoted_items[:-1])

    if len(quoted_items) > 2:
        humanized += ","

    return "{} {} {}".format(humanized, conjunction, quoted_items[-1])


def pluralize(container: Sized, if_one: str, if_multiple: str) -> str:
    if len(container) == 1:
        return if_one
    else:
        return if_multiple



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import shlex
from typing import Dict, Sequence

from .. import errors
from .._base_provider import Provider
from ._instance_info import InstanceInfo
from ._multipass_command import MultipassCommand
from snapcraft.internal.errors import SnapcraftEnvironmentError


logger = logging.getLogger(__name__)


def _get_stop_time() -> int:
    try:
        timeout = int(os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME", "10"))
    except ValueError:
        raise SnapcraftEnvironmentError(
            "'SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME' is incorrectly set, found {!r} "
            "but expected a number representing the minutes to delay the actual stop "
            "operation (or 0 to stop immediately).".format(
                os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT_STOP_TIME")
            )
        )

    return timeout


class _MachineSetting:
    def __init__(self, *, envvar: str, default: str) -> None:
        self._default = default
        self._envvar = envvar

    def get_value(self):
        value = os.getenv(self._envvar, self._default)
        if value != self._default:
            logger.warning(
                "{!r} was set to {!r} in the environment. "
                "Changing the default allocation upon user request".format(
                    self._envvar, value
                )
            )
        return value


class Multipass(Provider):
    """A multipass provider for snapcraft to execute its lifecycle."""

    @classmethod
    def _get_provider_name(cls):
        return "multipass"

    def _run(self, command: Sequence[str], hide_output: bool = False) -> None:
        command = ["sudo", "-i"] + list(command)
        self._multipass_cmd.execute(
            instance_name=self.instance_name, command=command, hide_output=hide_output
        )

    def _get_disk_image(self) -> str:
        if self.project.info.base is None:
            image = "snapcraft:core16"
        else:
            image = "snapcraft:{}".format(self.project.info.base)

        return image

    def _launch(self) -> None:
        cloud_user_data_filepath = self._get_cloud_user_data()
        image = self._get_disk_image()

        cpus = _MachineSetting(envvar="SNAPCRAFT_BUILD_ENVIRONMENT_CPU", default="2")
        mem = _MachineSetting(envvar="SNAPCRAFT_BUILD_ENVIRONMENT_MEMORY", default="2G")
        disk = _MachineSetting(
            envvar="SNAPCRAFT_BUILD_ENVIRONMENT_DISK", default="256G"
        )

        self._multipass_cmd.launch(
            instance_name=self.instance_name,
            cpus=cpus.get_value(),
            mem=mem.get_value(),
            disk=disk.get_value(),
            image=image,
            cloud_init=cloud_user_data_filepath,
        )

    def _start(self):
        try:
            self._get_instance_info()
        except errors.ProviderInfoError as instance_error:
            # Until we have proper multipass error codes to know if this
            # was a communication error we should keep this error tracking
            # and generation here.
            raise errors.ProviderInstanceNotFoundError(
                instance_name=self.instance_name
            ) from instance_error

        self._multipass_cmd.start(instance_name=self.instance_name)

    def _mount(
        self,
        *,
        mountpoint: str,
        dev_or_path: str,
        uid_map: Dict[str, str] = None,
        gid_map: Dict[str, str] = None
    ) -> None:
        target = "{}:{}".format(self.instance_name, mountpoint)
        self._multipass_cmd.mount(
            source=dev_or_path, target=target, uid_map=uid_map, gid_map=gid_map
        )

    def _umount(self, *, mountpoint: str) -> None:
        mount = "{}:{}".format(self.instance_name, mountpoint)
        self._multipass_cmd.umount(mount=mount)

    def _mount_snaps_directory(self) -> None:
        # https://github.com/snapcore/snapd/blob/master/dirs/dirs.go
        # CoreLibExecDir
        path = os.path.join(os.path.sep, "var", "lib", "snapd", "snaps")
        self._mount(mountpoint=self._SNAPS_MOUNTPOINT, dev_or_path=path)

    def _unmount_snaps_directory(self):
        self._umount(mountpoint=self._SNAPS_MOUNTPOINT)

    def _push_file(self, *, source: str, destination: str) -> None:
        destination = "{}:{}".format(self.instance_name, destination)
        self._multipass_cmd.copy_files(source=source, destination=destination)

    def __init__(self, *, project, echoer, is_ephemeral: bool = False) -> None:
        super().__init__(project=project, echoer=echoer, is_ephemeral=is_ephemeral)
        self._multipass_cmd = MultipassCommand()
        self._instance_info = None  # type: InstanceInfo

    def create(self) -> None:
        """Create the multipass instance and setup the build environment."""
        self.launch_instance()
        self._instance_info = self._get_instance_info()

    def destroy(self) -> None:
        """Destroy the instance, trying to stop it first."""
        try:
            instance_info = self._instance_info = self._get_instance_info()
        except errors.ProviderInfoError:
            return

        if instance_info.is_stopped():
            return

        stop_time = _get_stop_time()
        if stop_time > 0:
            try:
                self._multipass_cmd.stop(
                    instance_name=self.instance_name, time=stop_time
                )
            except errors.ProviderStopError:
                self._multipass_cmd.stop(instance_name=self.instance_name)
        else:
            self._multipass_cmd.stop(instance_name=self.instance_name)

        if self._is_ephemeral:
            self.clean_project()

    def mount_project(self) -> None:
        # Resolve the home directory
        home_dir = (
            self._multipass_cmd.execute(
                command=["sudo", "-i", "printenv", "HOME"],
                hide_output=True,
                instance_name=self.instance_name,
            )
            .decode()
            .strip()
        )
        project_mountpoint = os.path.join(home_dir, "project")

        # multipass keeps the mount active, so check if it is there first.
        if not self._instance_info.is_mounted(project_mountpoint):
            self._mount(
                mountpoint=project_mountpoint,
                dev_or_path=self.project._project_dir,
                uid_map={str(os.getuid()): "0"},
                gid_map={str(os.getgid()): "0"},
            )

    def provision_project(self, tarball: str) -> None:
        """Provision the multipass instance with the project to work with."""
        # TODO add instance check.
        # Step 0, sanitize the input
        tarball = shlex.quote(tarball)

        # First create a working directory
        self._multipass_cmd.execute(
            command=["sudo", "-i", "mkdir", self._INSTANCE_PROJECT_DIR],
            instance_name=self.instance_name,
        )

        # Then copy the tarball over
        destination = "{}:{}".format(self.instance_name, tarball)
        self._multipass_cmd.copy_files(source=tarball, destination=destination)

        # Finally extract it into project_dir.
        extract_cmd = [
            "sudo",
            "-i",
            "tar",
            "-xvf",
            tarball,
            "-C",
            self._INSTANCE_PROJECT_DIR,
        ]
        self._multipass_cmd.execute(
            command=extract_cmd, instance_name=self.instance_name
        )

    def clean_project(self) -> bool:
        was_cleaned = super().clean_project()
        if was_cleaned:
            self._multipass_cmd.delete(instance_name=self.instance_name, purge=True)
        return was_cleaned

    def build_project(self) -> None:
        # TODO add instance check.
        self._multipass_cmd.execute(
            command=["sudo", "-i", "snapcraft", "snap", "--output", self.snap_filename],
            instance_name=self.instance_name,
        )

    def retrieve_snap(self) -> str:
        # TODO add instance check.
        source = "{}:{}/{}".format(
            self.instance_name, self._INSTANCE_PROJECT_DIR, self.snap_filename
        )
        self._multipass_cmd.copy_files(source=source, destination=self.snap_filename)
        return self.snap_filename

    def shell(self) -> None:
        self._multipass_cmd.execute(
            instance_name=self.instance_name, command=["sudo", "-i", "/bin/bash"]
        )

    def _get_instance_info(self):
        instance_info_raw = self._multipass_cmd.info(
            instance_name=self.instance_name, output_format="json"
        )
        return InstanceInfo.from_json(
            instance_name=self.instance_name, json_info=instance_info_raw.decode()
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The meson plugin is useful for building meson based parts.

Meson based projects are projects that have a meson.build that drives the
build.

This plugin always runs 'meson snapbuild' followed by 'ninja' and
'ninja install'.

Additionally, this plugin uses the following plugin-specific keywords:

    - meson-version:
      (string)
      The version of meson to install from PyPI.
      If unspecified, the latest released version of meson will be used.
    - meson-parameters:
      (list of strings)
      Pass the given parameters to the meson command.

"""

import subprocess

import os
import snapcraft

from snapcraft.internal import errors


class MesonPlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["meson-parameters"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["meson-version"] = {"type": "string", "default": ""}
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        return ["meson-version"]

    @classmethod
    def get_build_properties(cls):
        return ["meson-parameters"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_base_tools(project.info.base)

        self.snapbuildname = "snapbuild"
        self.mesonbuilddir = os.path.join(self.builddir, self.snapbuildname)

    def _setup_base_tools(self, base):
        if base in ("core16", "core18"):
            self.build_packages.append("python3-pip")
            self.build_packages.append("python3-setuptools")
            self.build_packages.append("python3-wheel")
            self.build_packages.append("ninja-build")
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)
        if base == "core18":
            self.build_packages.append("python3-distutils")

    def pull(self):
        super().pull()

        if self.options.meson_version:
            meson_package = "meson=={}".format(self.options.meson_version)
        else:
            meson_package = "meson"

        try:
            subprocess.check_call(
                ["python3", "-m", "pip", "install", "-U", meson_package]
            )
        except subprocess.CalledProcessError as call_error:
            raise errors.SnapcraftPluginCommandError(
                command=call_error.cmd,
                part_name=self.name,
                exit_code=call_error.returncode,
            ) from call_error

    def build(self):
        super().build()
        self._run_meson()
        self._run_ninja_build_default()
        self._run_ninja_install()

    def _run_meson(self):
        os.makedirs(self.mesonbuilddir, exist_ok=True)
        meson_command = ["meson"]
        if self.options.meson_parameters:
            meson_command.extend(self.options.meson_parameters)
        meson_command.append(self.snapbuildname)
        self.run(meson_command)

    def _run_ninja_build_default(self):
        ninja_command = ["ninja"]
        self.run(ninja_command, cwd=self.mesonbuilddir)

    def _run_ninja_install(self):
        env = os.environ.copy()
        env["DESTDIR"] = self.installdir
        ninja_install_command = ["ninja", "install"]
        self.run(ninja_install_command, env=env, cwd=self.mesonbuilddir)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This plugin is useful for building parts that use gradle.

The gradle build system is commonly used to build Java projects.
The plugin requires a pom.xml in the root of the source tree.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - gradle-options:
      (list of strings)
      Flags to pass to the build using the gradle semantics for parameters.
      The 'jar' option is always passed in as the last parameter.

    - gradle-output-dir:
      (string; default: 'build/libs')
      The output directory where the resulting jar or war files from gradle[w]
      are generated.

    - gradle-version:
      (string)
      The version of gradle you want to use to build the source artifacts.
      Defaults to the current release downloadable from
      https://services.gradle.org/distributions/
      The entry is ignored if gradlew is found.

    - gradle-version-checksum:
      (string)
      The checksum for gradle-version in the form of <digest-type>/<digest>.
      As an example "sha512/2a803f578f341e164f6753e410413d16ab60fab...".

    - gradle-openjdk-version:
      (string)
      openjdk version available to the base to use. If not set the latest
      version available to the base will be used.
"""

import logging
import os
import urllib.parse
from glob import glob
from typing import Sequence

import snapcraft
from snapcraft import file_utils, formatting_utils
from snapcraft.internal import errors, sources

logger = logging.getLogger(__name__)


_DEFAULT_GRADLE_VERSION = "4.10.2"
_DEFAULT_GRADLE_CHECKSUM = (
    "sha256/b49c6da1b2cb67a0caf6c7480630b51c70a11ca2016ff2f555eaeda863143a29"
)
_GRADLE_URL = "https://services.gradle.org/distributions/gradle-{version}-bin.zip"


class UnsupportedJDKVersionError(errors.SnapcraftError):

    fmt = (
        "The gradle-openjdk-version plugin property was set to {version!r}.\n"
        "Valid values for the {base!r} base are: {valid_versions}."
    )

    def __init__(
        self, *, base: str, version: str, valid_versions: Sequence[str]
    ) -> None:
        super().__init__(
            base=base,
            version=version,
            valid_versions=formatting_utils.humanize_list(
                valid_versions, conjunction="or"
            ),
        )


class GradlePlugin(snapcraft.BasePlugin):
    @classmethod
    def schema(cls):
        schema = super().schema()
        schema["properties"]["gradle-options"] = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }
        schema["properties"]["gradle-output-dir"] = {
            "type": "string",
            "default": "build/libs",
        }

        schema["properties"]["gradle-version"] = {"type": "string"}

        schema["properties"]["gradle-version-checksum"] = {"type": "string"}

        schema["properties"]["gradle-openjdk-version"] = {
            "type": "string",
            "default": "",
        }

        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the pull step dirty.
        return ["gradle-version", "gradle-version-checksum", "gradle-openjdk-version"]

    @classmethod
    def get_build_properties(cls):
        # Inform Snapcraft of the properties associated with building. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return super().get_build_properties() + ["gradle-options", "gradle-output-dir"]

    @property
    def _gradle_tar(self):
        if self._gradle_tar_handle is None:
            gradle_uri = _GRADLE_URL.format(version=self._gradle_version)
            self._gradle_tar_handle = sources.Zip(
                gradle_uri, self._gradle_dir, source_checksum=self._gradle_checksum
            )
        return self._gradle_tar_handle

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._setup_gradle()
        self._setup_base_tools(project.info.base)

    def _setup_base_tools(self, base):
        if base not in ("core16", "core18"):
            raise errors.PluginBaseError(
                part_name=self.name, base=self.project.info.base
            )

        if base == "core16":
            valid_versions = ["8", "9"]
        elif base == "core18":
            valid_versions = ["8", "11"]

        version = self.options.gradle_openjdk_version
        if version and version not in valid_versions:
            raise UnsupportedJDKVersionError(
                version=version, base=base, valid_versions=valid_versions
            )
        elif not version:
            # Get the latest version from the slice
            version = valid_versions[-1]

        self.stage_packages.append("openjdk-{}-jre-headless".format(version))
        self.build_packages.append("openjdk-{}-jdk-headless".format(version))
        self.build_packages.append("ca-certificates-java")

    def _using_gradlew(self) -> bool:
        return os.path.isfile(os.path.join(self.sourcedir, "gradlew"))

    def _setup_gradle(self):
        self._gradle_tar_handle = None
        self._gradle_dir = os.path.join(self.partdir, "gradle")
        if self.options.gradle_version:
            self._gradle_version = self.options.gradle_version
            self._gradle_checksum = self.options.gradle_version_checksum
        else:
            self._gradle_version = _DEFAULT_GRADLE_VERSION
            self._gradle_checksum = _DEFAULT_GRADLE_CHECKSUM

    def pull(self):
        super().pull()

        if self._using_gradlew():
            logger.info("Found gradlew, skipping gradle setup.")
            return

        os.makedirs(self._gradle_dir, exist_ok=True)
        self._gradle_tar.download()

    def build(self):
        super().build()

        if self._using_gradlew():
            gradle_cmd = ["./gradlew"]
        else:
            self._gradle_tar.provision(self._gradle_dir, keep_zip=True)
            gradle_cmd = ["gradle"]
        self.run(
            gradle_cmd
            + self._get_proxy_options()
            + self.options.gradle_options
            + ["jar"],
            rootdir=self.builddir,
        )

        src = os.path.join(self.builddir, self.options.gradle_output_dir)
        jarfiles = glob(os.path.join(src, "*.jar"))
        warfiles = glob(os.path.join(src, "*.war"))

        if len(jarfiles) > 0:
            basedir = "jar"
        elif len(warfiles) > 0:
            basedir = "war"
            jarfiles = warfiles
        else:
            raise RuntimeError("Could not find any built jar files for part")

        file_utils.link_or_copy_tree(
            src,
            os.path.join(self.installdir, basedir),
            copy_function=lambda src, dst: file_utils.link_or_copy(
                src, dst, self.installdir
            ),
        )

        self._create_symlinks()

    def _create_symlinks(self):
        if self.project.info.base not in ("core18", "core19"):
            raise errors.PluginBaseError(
                part_name=self.name, base=self.project.info.base
            )

        os.makedirs(os.path.join(self.installdir, "bin"), exist_ok=True)
        java_bin = glob(
            os.path.join(
                self.installdir,
                "usr",
                "lib",
                "jvm",
                "java-{}-openjdk-*".format(self.options.gradle_openjdk_version),
                "jre",
                "bin",
                "java",
            )
        )[0]
        os.symlink(
            os.path.relpath(java_bin, os.path.join(self.installdir, "bin")),
            os.path.join(self.installdir, "bin", "java"),
        )

    def run(self, cmd, rootdir):
        super().run(cmd, cwd=rootdir, env=self._build_environment())

    def _build_environment(self):
        if self._using_gradlew():
            return

        env = os.environ.copy()
        gradle_bin = os.path.join(
            self._gradle_dir, "gradle-{}".format(self._gradle_version), "bin"
        )
        print(gradle_bin)

        if env.get("PATH"):
            new_path = "{}:{}".format(gradle_bin, env.get("PATH"))
        else:
            new_path = gradle_bin

        env["PATH"] = new_path
        return env

    def _get_proxy_options(self):
        proxy_options = []
        for var in ("http", "https"):
            proxy = os.environ.get("{}_proxy".format(var), False)
            if proxy:
                parsed_url = urllib.parse.urlparse(proxy)
                proxy_options.append(
                    "-D{}.proxyHost={}".format(var, parsed_url.hostname)
                )
                if parsed_url.port:
                    proxy_options.append(
                        "-D{}.proxyPort={}".format(var, parsed_url.port)
                    )
                if parsed_url.username:
                    proxy_options.append(
                        "-D{}.proxyUser={}".format(var, parsed_url.username)
                    )
                if parsed_url.password:
                    proxy_options.append(
                        "-D{}.proxyPassword={}".format(var, parsed_url.password)
                    )
        return proxy_options



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import collections
import contextlib
import itertools
import logging
import os
import re
import shlex
import shutil
import stat
import subprocess
from typing import Any, Dict, List, Set  # noqa

from snapcraft import file_utils, formatting_utils, yaml_utils
from snapcraft import shell_utils
from snapcraft.project import _schema
from snapcraft.internal import common, errors, project_loader
from snapcraft.internal.project_loader import _config
from snapcraft.extractors import _metadata
from snapcraft.internal.deprecations import handle_deprecation_notice
from snapcraft.internal.meta import (
    _desktop,
    _errors as meta_errors,
    _manifest,
    _version,
)


logger = logging.getLogger(__name__)


_MANDATORY_PACKAGE_KEYS = ["name", "version", "summary", "description"]

_OPTIONAL_PACKAGE_KEYS = [
    "architectures",
    "assumes",
    "base",
    "confinement",
    "environment",
    "epoch",
    "grade",
    "hooks",
    "layout",
    "license",
    "plugs",
    "slots",
    "title",
    "type",
]


class OctInt(yaml_utils.SnapcraftYAMLObject):
    """An int represented in octal form."""

    yaml_tag = u"!OctInt"

    def __init__(self, value):
        super().__init__()
        self._value = value

    @classmethod
    def to_yaml(cls, dumper, data):
        """
        Convert a Python object to a representation node.
        """
        return dumper.represent_scalar(
            "tag:yaml.org,2002:int", "{:04o}".format(data._value)
        )


# From snapd's snap/validate.go, appContentWhitelist, with a slight modification: don't
# allow leading slashes.
_APP_COMMAND_PATTERN = re.compile("^[A-Za-z0-9. _#:$-][A-Za-z0-9/. _#:$-]*$")


def create_snap_packaging(project_config: _config.Config) -> str:
    """Create snap.yaml and related assets in meta.

    Create the meta directory and provision it with snap.yaml in the snap dir
    using information from config_data. Also copy in the local 'snap'
    directory, and generate wrappers for hooks coming from parts.

    :param dict config_data: project values defined in snapcraft.yaml.
    :return: meta_dir.
    """

    # Update config_data using metadata extracted from the project
    _update_yaml_with_extracted_metadata(project_config.data, project_config.parts)

    # Now that we've updated config_data with random stuff extracted from
    # parts, re-validate it to ensure the it still conforms with the schema.
    validator = _schema.Validator(project_config.data)
    validator.validate(source="properties")

    # Update default values
    _update_yaml_with_defaults(project_config.data, project_config.validator.schema)

    # Ensure the YAML contains all required keywords before continuing to
    # use it to generate the snap.yaml.
    _ensure_required_keywords(project_config.data)

    packaging = _SnapPackaging(project_config)
    packaging.write_snap_yaml()
    packaging.setup_assets()
    packaging.generate_hook_wrappers()
    packaging.write_snap_directory()

    return packaging.meta_dir


def _update_yaml_with_extracted_metadata(
    config_data: Dict[str, Any], parts_config: project_loader.PartsConfig
) -> None:
    if "adopt-info" in config_data:
        part_name = config_data["adopt-info"]
        part = parts_config.get_part(part_name)
        if not part:
            raise meta_errors.AdoptedPartMissingError(part_name)

        pull_state = part.get_pull_state()
        build_state = part.get_build_state()
        stage_state = part.get_stage_state()
        prime_state = part.get_prime_state()

        # Get the metadata from the pull step first.
        metadata = pull_state.extracted_metadata["metadata"]

        # Now update it using the metadata from the build step (i.e. the data
        # from the build step takes precedence over the pull step).
        metadata.update(build_state.extracted_metadata["metadata"])

        # Now make sure any scriptlet data are taken into account. Later steps
        # take precedence, and scriptlet data (even in earlier steps) take
        # precedence over extracted data.
        metadata.update(pull_state.scriptlet_metadata)
        metadata.update(build_state.scriptlet_metadata)
        metadata.update(stage_state.scriptlet_metadata)
        metadata.update(prime_state.scriptlet_metadata)

        if not metadata:
            # If we didn't end up with any metadata, let's ensure this part was
            # actually supposed to parse info. If not, let's try to be very
            # clear about what's happening, here. We do this after checking for
            # metadata because metadata could be supplied by scriptlets, too.
            if "parse-info" not in config_data["parts"][part_name]:
                raise meta_errors.AdoptedPartNotParsingInfo(part_name)

        _adopt_info(config_data, metadata)


def _adopt_info(
    config_data: Dict[str, Any], extracted_metadata: _metadata.ExtractedMetadata
):
    ignored_keys = _adopt_keys(config_data, extracted_metadata)
    if ignored_keys:
        logger.warning(
            "The {keys} {plural_property} {plural_is} specified in adopted "
            "info as well as the YAML: taking the {plural_property} from the "
            "YAML".format(
                keys=formatting_utils.humanize_list(list(ignored_keys), "and"),
                plural_property=formatting_utils.pluralize(
                    ignored_keys, "property", "properties"
                ),
                plural_is=formatting_utils.pluralize(ignored_keys, "is", "are"),
            )
        )


def _adopt_keys(
    config_data: Dict[str, Any], extracted_metadata: _metadata.ExtractedMetadata
) -> Set[str]:
    ignored_keys = set()
    metadata_dict = extracted_metadata.to_dict()

    # desktop_file_paths and common_ids are special cases that will be handled
    # after all the top level snapcraft.yaml keys.
    ignore = ("desktop_file_paths", "common_id")
    overrides = ((k, v) for k, v in metadata_dict.items() if k not in ignore)

    for key, value in overrides:
        if key not in config_data:
            if key == "icon":
                if _icon_file_exists() or not os.path.exists(str(value)):
                    # Do not overwrite the icon file.
                    continue
            config_data[key] = value
        else:
            ignored_keys.add(key)

    if "desktop_file_paths" in metadata_dict and "common_id" in metadata_dict:
        app_name = _get_app_name_from_common_id(
            config_data, str(metadata_dict["common_id"])
        )
        if app_name and not _desktop_file_exists(app_name):
            for desktop_file_path in metadata_dict["desktop_file_paths"]:
                if os.path.exists(desktop_file_path):
                    config_data["apps"][app_name]["desktop"] = desktop_file_path
                    break

    return ignored_keys


def _icon_file_exists() -> bool:
    """Check if the icon is specified as a file in the assets dir.

    The icon file can be in two directories: 'setup/gui' (deprecated) or
    'snap/gui'. The icon can be named 'icon.png' or 'icon.svg'.

    :returns: True if the icon file exists, False otherwise.
    """
    for icon_path in (
        os.path.join(asset_dir, "gui", icon_file)
        for (asset_dir, icon_file) in itertools.product(
            ["setup", "snap"], ["icon.png", "icon.svg"]
        )
    ):
        if os.path.exists(icon_path):
            return True
    else:
        return False


def _get_app_name_from_common_id(config_data: Dict[str, Any], common_id: str) -> str:
    """Get the snap app name with a common-id.

    :params dict config_data: Project values defined in snapcraft.yaml.
    :params common_id: The common identifier across multiple packaging
        formats.
    :returns: The name of the snap app with the common-id.

    """
    if "apps" in config_data:
        for app in config_data["apps"]:
            if config_data["apps"][app].get("common-id") == common_id:
                return app
    return None


def _desktop_file_exists(app_name: str) -> bool:
    """Check if the desktop file is specified as a file in the assets dir.

    The desktop file can be in two directories: 'setup/gui' (deprecated) or
    'snap/gui'.

    :params app_name: The name of the snap app.
    :returns: True if the desktop file exists, False otherwise.
    """
    for desktop_path in (
        os.path.join(asset_dir, "gui", "{}.desktop".format(app_name))
        for asset_dir in ["setup", "snap"]
    ):
        if os.path.exists(desktop_path):
            return True
    else:
        return False


def _update_yaml_with_defaults(config_data, schema):
    # Ensure that grade and confinement have their defaults applied, if
    # necessary. Defaults are taken from the schema. Technically these are the
    # only two optional keywords currently WITH defaults, but we don't want to
    # risk setting something that we add later on accident.
    for key in ("confinement", "grade"):
        if key not in config_data:
            with contextlib.suppress(KeyError):
                default = schema[key]["default"]
                config_data[key] = default
                logger.warn(
                    "{!r} property not specified: defaulting to {!r}".format(
                        key, default
                    )
                )

    # Set default adapter
    app_schema = schema["apps"]["patternProperties"]["^[a-zA-Z0-9](?:-?[a-zA-Z0-9])*$"][
        "properties"
    ]
    default_adapter = app_schema["adapter"]["default"]
    for app in config_data.get("apps", {}).values():
        if "adapter" not in app:
            app["adapter"] = default_adapter


def _ensure_required_keywords(config_data):
    # Verify that all mandatory keys have been satisfied
    missing_keys = []  # type: List[str]
    for key in _MANDATORY_PACKAGE_KEYS:
        if key not in config_data:
            missing_keys.append(key)

    if missing_keys:
        raise meta_errors.MissingSnapcraftYamlKeysError(keys=missing_keys)


class _SnapPackaging:
    @property
    def meta_dir(self) -> str:
        return self._meta_dir

    def __init__(self, project_config: _config.Config) -> None:
        self._project_config = project_config
        self._snapcraft_yaml_path = project_config.project.info.snapcraft_yaml_file_path
        self._prime_dir = project_config.project.prime_dir
        self._parts_dir = project_config.project.parts_dir
        self._arch_triplet = project_config.project.arch_triplet
        self._global_state_file = project_config.project._global_state_file
        self._is_host_compatible_with_base = (
            project_config.project.is_host_compatible_with_base
        )
        self._meta_dir = os.path.join(self._prime_dir, "meta")
        self._config_data = project_config.data.copy()
        self._original_snapcraft_yaml = project_config.project.info.get_raw_snapcraft()
        self._meta_runner = os.path.join(
            self._prime_dir, "snap", "command-chain", "snapcraft-runner"
        )

        self._install_path_pattern = re.compile(
            r"{}/[a-z0-9][a-z0-9+-]*/install".format(re.escape(self._parts_dir))
        )

        os.makedirs(self._meta_dir, exist_ok=True)

    def write_snap_yaml(self) -> str:
        common.env = self._project_config.snap_env()

        try:
            # Only generate the meta command chain if there are apps that need it
            if self._config_data.get("apps", None):
                self._generate_command_chain()

            package_snap_path = os.path.join(self.meta_dir, "snap.yaml")
            snap_yaml = self._compose_snap_yaml()

            with open(package_snap_path, "w") as f:
                yaml_utils.dump(snap_yaml, stream=f)

            return snap_yaml
        finally:
            common.reset_env()

    def setup_assets(self) -> None:
        # We do _setup_from_setup first since it is legacy and let the
        # declarative items take over.
        self._setup_gui()

        if "icon" in self._config_data:
            # TODO: use developer.ubuntu.com once it has updated documentation.
            icon_ext = self._config_data["icon"].split(os.path.extsep)[-1]
            icon_dir = os.path.join(self.meta_dir, "gui")
            icon_path = os.path.join(icon_dir, "icon.{}".format(icon_ext))
            if not os.path.exists(icon_dir):
                os.mkdir(icon_dir)
            if os.path.exists(icon_path):
                os.unlink(icon_path)
            file_utils.link_or_copy(self._config_data["icon"], icon_path)

        if self._config_data.get("type", "") == "gadget":
            if not os.path.exists("gadget.yaml"):
                raise errors.MissingGadgetError()
            file_utils.link_or_copy(
                "gadget.yaml", os.path.join(self.meta_dir, "gadget.yaml")
            )

    def _generate_command_chain(self):
        # Classic confinement or building on a host that does not match the target base
        # means we cannot setup an environment that will work.
        if (
            self._config_data["confinement"] == "classic"
            or not self._is_host_compatible_with_base
        ):
            assembled_env = None
        else:
            assembled_env = common.assemble_env()
            assembled_env = assembled_env.replace(self._prime_dir, "$SNAP")
            assembled_env = self._install_path_pattern.sub("$SNAP", assembled_env)

            if assembled_env:
                os.makedirs(os.path.dirname(self._meta_runner), exist_ok=True)
                with open(self._meta_runner, "w") as f:
                    print("#!/bin/sh", file=f)
                    print(assembled_env, file=f)
                    print(
                        "export LD_LIBRARY_PATH=$SNAP_LIBRARY_PATH:$LD_LIBRARY_PATH",
                        file=f,
                    )
                    print('exec "$@"', file=f)
                os.chmod(self._meta_runner, 0o755)

    def _record_manifest_and_source_snapcraft_yaml(self):
        prime_snap_dir = os.path.join(self._prime_dir, "snap")
        recorded_snapcraft_yaml_path = os.path.join(prime_snap_dir, "snapcraft.yaml")
        if os.path.isfile(recorded_snapcraft_yaml_path):
            os.unlink(recorded_snapcraft_yaml_path)
        manifest_file_path = os.path.join(prime_snap_dir, "manifest.yaml")
        if os.path.isfile(manifest_file_path):
            os.unlink(manifest_file_path)

        # FIXME hide this functionality behind a feature flag for now
        if os.environ.get("SNAPCRAFT_BUILD_INFO"):
            os.makedirs(prime_snap_dir, exist_ok=True)
            shutil.copy2(self._snapcraft_yaml_path, recorded_snapcraft_yaml_path)
            annotated_snapcraft = _manifest.annotate_snapcraft(
                copy.deepcopy(self._config_data),
                self._parts_dir,
                self._global_state_file,
            )
            with open(manifest_file_path, "w") as manifest_file:
                yaml_utils.dump(annotated_snapcraft, stream=manifest_file)

    def write_snap_directory(self) -> None:
        # First migrate the snap directory. It will overwrite any conflicting
        # files.
        for root, directories, files in os.walk("snap"):
            with contextlib.suppress(ValueError):
                directories.remove(".snapcraft")
            with contextlib.suppress(ValueError):
                # The snapcraft.yaml is migrated later
                files.remove("snapcraft.yaml")

            for directory in directories:
                source = os.path.join(root, directory)
                destination = os.path.join(self._prime_dir, source)
                file_utils.create_similar_directory(source, destination)

            for file_path in files:
                source = os.path.join(root, file_path)
                destination = os.path.join(self._prime_dir, source)
                with contextlib.suppress(FileNotFoundError):
                    os.remove(destination)
                file_utils.link_or_copy(source, destination)

        # Now copy the assets contained within the snap directory directly into
        # meta.
        for origin in ["gui", "hooks"]:
            src_dir = os.path.join("snap", origin)
            dst_dir = os.path.join(self.meta_dir, origin)
            if os.path.isdir(src_dir):
                os.makedirs(dst_dir, exist_ok=True)
                for asset in os.listdir(src_dir):
                    source = os.path.join(src_dir, asset)
                    destination = os.path.join(dst_dir, asset)

                    # First, verify that the hook is actually executable
                    if origin == "hooks":
                        _validate_hook(source)

                    with contextlib.suppress(FileNotFoundError):
                        os.remove(destination)

                    file_utils.link_or_copy(source, destination)

        self._record_manifest_and_source_snapcraft_yaml()

    def generate_hook_wrappers(self) -> None:
        snap_hooks_dir = os.path.join(self._prime_dir, "snap", "hooks")
        hooks_dir = os.path.join(self._prime_dir, "meta", "hooks")
        if os.path.isdir(snap_hooks_dir):
            os.makedirs(hooks_dir, exist_ok=True)
            for hook_name in os.listdir(snap_hooks_dir):
                file_path = os.path.join(snap_hooks_dir, hook_name)
                # First, verify that the hook is actually executable
                if not os.stat(file_path).st_mode & stat.S_IEXEC:
                    raise meta_errors.CommandError(
                        "hook {!r} is not executable".format(hook_name)
                    )

                hook_exec = os.path.join("$SNAP", "snap", "hooks", hook_name)
                hook_path = os.path.join(hooks_dir, hook_name)
                with contextlib.suppress(FileNotFoundError):
                    os.remove(hook_path)

                self._write_wrap_exe(hook_exec, hook_path)

    def _setup_gui(self):
        # Handles the setup directory which only contains gui assets.
        setup_dir = "setup"
        if not os.path.exists(setup_dir):
            return

        handle_deprecation_notice("dn3")

        gui_src = os.path.join(setup_dir, "gui")
        gui_dst = os.path.join(self.meta_dir, "gui")
        if os.path.exists(gui_src):
            for f in os.listdir(gui_src):
                if not os.path.exists(gui_dst):
                    os.mkdir(gui_dst)
                shutil.copy2(os.path.join(gui_src, f), gui_dst)

    def _compose_snap_yaml(self):
        """Create a new dictionary from config_data to obtain snap.yaml.

        Missing key exceptions will be raised if config_data does not contain
        all the _MANDATORY_PACKAGE_KEYS, config_data can be validated against
        the snapcraft schema.

        Keys that are in _OPTIONAL_PACKAGE_KEYS are ignored if not there.
        """
        snap_yaml = collections.OrderedDict()

        for key_name in _MANDATORY_PACKAGE_KEYS:
            snap_yaml[key_name] = self._config_data[key_name]

        # Reparse the version, the order should stick.
        snap_yaml["version"] = _version.get_version(
            self._config_data["version"], self._config_data.get("version-script")
        )

        for key_name in _OPTIONAL_PACKAGE_KEYS:
            if key_name in self._config_data:
                snap_yaml[key_name] = self._config_data[key_name]

        if "apps" in self._config_data:
            _verify_app_paths(basedir=self._prime_dir, apps=self._config_data["apps"])
            snap_yaml["apps"] = self._wrap_apps(self._config_data["apps"])
            self._render_socket_modes(snap_yaml["apps"])
            self._validate_command_chain(snap_yaml["apps"])

        self._process_passthrough_properties(snap_yaml)
        assumes = _determine_assumes(self._config_data)
        if assumes:
            # Sorting for consistent results (order doesn't matter)
            snap_yaml["assumes"] = sorted(set(snap_yaml.get("assumes", [])) | assumes)

        return snap_yaml

    def _write_wrap_exe(self, wrapexec, wrappath, shebang=None, args=None, cwd=None):
        if args:
            quoted_args = ['"{}"'.format(arg) for arg in args]
        else:
            quoted_args = []
        args = " ".join(quoted_args) + ' "$@"' if args else '"$@"'
        cwd = "cd {}".format(cwd) if cwd else ""

        executable = '"{}"'.format(wrapexec)

        if shebang:
            if shebang.startswith("/usr/bin/env "):
                shebang = shell_utils.which(shebang.split()[1])
            new_shebang = self._install_path_pattern.sub("$SNAP", shebang)
            new_shebang = re.sub(self._prime_dir, "$SNAP", new_shebang)
            if new_shebang != shebang:
                # If the shebang was pointing to and executable within the
                # local 'parts' dir, have the wrapper script execute it
                # directly, since we can't use $SNAP in the shebang itself.
                executable = '"{}" "{}"'.format(new_shebang, wrapexec)

        with open(wrappath, "w+") as f:
            print("#!/bin/sh", file=f)
            if cwd:
                print("{}".format(cwd), file=f)
            print("exec {} {}".format(executable, args), file=f)

        os.chmod(wrappath, 0o755)

    def _wrap_exe(self, command, basename=None):
        execparts = shlex.split(command)
        exepath = os.path.join(self._prime_dir, execparts[0])
        if basename:
            wrappath = os.path.join(self._prime_dir, basename) + ".wrapper"
        else:
            wrappath = exepath + ".wrapper"
        shebang = None

        if os.path.exists(wrappath):
            os.remove(wrappath)

        wrapexec = "$SNAP/{}".format(execparts[0])
        if not os.path.exists(exepath) and "/" not in execparts[0]:
            _find_bin(execparts[0], self._prime_dir)
            wrapexec = execparts[0]
        else:
            with open(exepath, "rb") as exefile:
                # If the file has a she-bang, the path might be pointing to
                # the local 'parts' dir. Extract it so that _write_wrap_exe
                # will have a chance to rewrite it.
                if exefile.read(2) == b"#!":
                    shebang = exefile.readline().strip().decode("utf-8")

        self._write_wrap_exe(wrapexec, wrappath, shebang=shebang, args=execparts[1:])

        return os.path.relpath(wrappath, self._prime_dir)

    def _wrap_apps(self, apps: Dict[str, Any]) -> Dict[str, Any]:
        apps = copy.deepcopy(apps)
        gui_dir = os.path.join(self.meta_dir, "gui")
        if not os.path.exists(gui_dir):
            os.mkdir(gui_dir)
        for f in os.listdir(gui_dir):
            if os.path.splitext(f)[1] == ".desktop":
                os.remove(os.path.join(gui_dir, f))
        for app_name, app in apps.items():
            adapter = project_loader.Adapter[app.pop("adapter").upper()]
            if adapter == project_loader.Adapter.LEGACY:
                self._wrap_app(app_name, app)
            elif adapter == project_loader.Adapter.FULL:
                self._add_command_chain_to_app(app_name, app)
            self._generate_desktop_file(app_name, app)
        return apps

    def _add_command_chain_to_app(self, app_name, app):
        # First, validate the commands. Without a wrapper, we're bound by snapd's
        # restrictions. No quotes, and the first string must be a relative path from
        # the root of the snap.
        commands = (app[k] for k in ("command", "stop-command") if k in app)
        for command in commands:
            if not _APP_COMMAND_PATTERN.match(command):
                raise errors.InvalidAppCommandFormatError(command, app_name)

            command_without_args = command.split()[0]
            binary_path = os.path.join(self._prime_dir, command_without_args)
            is_executable = False
            with contextlib.suppress(FileNotFoundError):
                mode = os.stat(binary_path).st_mode
                is_executable = (
                    mode & stat.S_IXUSR or mode & stat.S_IXGRP or mode & stat.S_IXOTH
                )

            if not is_executable:
                raise errors.InvalidAppCommandError(command_without_args, app_name)

        # Finally, assuming there is a meta runner, add it to the BEGINNING of the
        # command chain. This is to ensure all subsequent commands in the chain get
        # a valid environment.
        if os.path.isfile(self._meta_runner):
            app["command-chain"] = [
                os.path.relpath(self._meta_runner, self._prime_dir)
            ] + app.get("command-chain", [])

    def _wrap_app(self, name, app):
        cmds = (k for k in ("command", "stop-command") if k in app)
        for k in cmds:
            try:
                new_command = self._wrap_exe(app[k], "{}-{}".format(k, name))
            except FileNotFoundError:
                raise errors.InvalidAppCommandError(command=app[k], app_name=app)
            except meta_errors.CommandError as e:
                raise errors.InvalidAppCommandError(str(e), name)

            if os.path.isfile(self._meta_runner):
                snap_runner = os.path.relpath(self._meta_runner, self._prime_dir)
                new_command = "{} $SNAP/{}".format(snap_runner, new_command)
            app[k] = new_command

    def _generate_desktop_file(self, name, app):
        desktop_file_name = app.pop("desktop", "")
        if desktop_file_name:
            desktop_file = _desktop.DesktopFile(
                name=name,
                filename=desktop_file_name,
                snap_name=self._config_data["name"],
                prime_dir=self._prime_dir,
            )
            desktop_file.parse_and_reformat()
            desktop_file.write(gui_dir=os.path.join(self.meta_dir, "gui"))

    def _render_socket_modes(self, apps: Dict[str, Any]) -> None:
        for app in apps.values():
            sockets = app.get("sockets", {})
            for socket in sockets.values():
                mode = socket.get("socket-mode")
                if mode is not None:
                    socket["socket-mode"] = OctInt(mode)

    def _validate_command_chain(self, apps: Dict[str, Any]) -> None:
        for app_name, app in apps.items():
            for item in app.get("command-chain", []):
                executable_path = os.path.join(self._prime_dir, item)

                # command-chain entries must always be relative to the root of the snap,
                # i.e. PATH is not used.
                if not _executable_is_valid(executable_path):
                    raise errors.InvalidCommandChainError(item, app_name)

    def _process_passthrough_properties(self, snap_yaml: Dict[str, Any]) -> None:
        passthrough_applied = False

        for section in ["apps", "hooks"]:
            if section in self._config_data:
                for name, value in snap_yaml[section].items():
                    if self._apply_passthrough(
                        value,
                        value.pop("passthrough", {}),
                        self._original_snapcraft_yaml[section][name],
                    ):
                        passthrough_applied = True

        if self._apply_passthrough(
            snap_yaml,
            self._config_data.get("passthrough", {}),
            self._original_snapcraft_yaml,
        ):
            passthrough_applied = True

        if passthrough_applied:
            logger.warn(
                "The 'passthrough' property is being used to "
                "propagate experimental properties to snap.yaml "
                "that have not been validated."
            )

    def _apply_passthrough(
        self,
        section: Dict[str, Any],
        passthrough: Dict[str, Any],
        original: collections.OrderedDict,
    ) -> bool:
        # Any value already in the original dictionary must
        # not be specified in passthrough at the same time.
        duplicates = list(original.keys() & passthrough.keys())
        if duplicates:
            raise meta_errors.AmbiguousPassthroughKeyError(duplicates)
        section.update(passthrough)
        return bool(passthrough)


def _determine_assumes(yaml_data: Dict[str, Any]) -> Set[str]:
    # Order doesn't really matter, but we don't want duplicates, so use a set
    assumes = set()

    # Check to see if any apps use the "full" adapter, which requires command-chain
    for app in yaml_data.get("apps", dict()).values():
        adapter = project_loader.Adapter[app["adapter"].upper()]
        if adapter == project_loader.Adapter.FULL:
            assumes.add("command-chain")

    return assumes


def _find_bin(binary, basedir):
    # If it doesn't exist it might be in the path
    logger.debug("Checking that {!r} is in the $PATH".format(binary))
    try:
        shell_utils.which(binary, cwd=basedir)
    except subprocess.CalledProcessError:
        raise meta_errors.CommandError(binary)


def _validate_hook(hook_path):
    if not os.stat(hook_path).st_mode & stat.S_IEXEC:
        asset = os.path.basename(hook_path)
        raise meta_errors.CommandError("hook {!r} is not executable".format(asset))


def _verify_app_paths(basedir, apps):
    for app in apps:
        path_entries = [i for i in ("desktop", "completer") if i in apps[app]]
        for path_entry in path_entries:
            file_path = os.path.join(basedir, apps[app][path_entry])
            if not os.path.exists(file_path):
                raise errors.SnapcraftPathEntryError(
                    app=app, key=path_entry, value=file_path
                )


def _executable_is_valid(path: str) -> bool:
    with contextlib.suppress(FileNotFoundError):
        mode = os.stat(path).st_mode
        return bool(mode & stat.S_IXUSR or mode & stat.S_IXGRP or mode & stat.S_IXOTH)

    return False



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import click

from snapcraft.project import Project, get_snapcraft_yaml


class HiddenOption(click.Option):
    def get_help_record(self, ctx):
        pass


_BUILD_OPTION_NAMES = [
    "--target-arch",
    "--debug",
    "--shell",
    "--shell-after",
    "--destructive-mode",
]

_BUILD_OPTIONS = [
    dict(metavar="<arch>", help="Target architecture to cross compile to"),
    dict(is_flag=True, help="Shells into the environment if the build fails."),
    dict(is_flag=True, help="Shells into the environment in lieu of the step to run."),
    dict(is_flag=True, help="Shells into the environment after the step has run."),
    dict(
        is_flag=True, help="Forces snapcraft to try and use the current host to build."
    ),
]


def add_build_options(hidden=False):
    def _add_build_options(func):
        for name, params in zip(
            reversed(_BUILD_OPTION_NAMES), reversed(_BUILD_OPTIONS)
        ):
            option = click.option(
                name, **params, cls=HiddenOption if hidden else click.Option
            )
            func = option(func)
        return func

    return _add_build_options


def get_project(
    *, is_managed_host: bool = False, skip_snapcraft_yaml: bool = False, **kwargs
):
    # We need to do this here until we can get_snapcraft_yaml as part of Project.
    if is_managed_host:
        os.chdir(os.path.expanduser(os.path.join("~", "project")))

    if skip_snapcraft_yaml:
        snapcraft_yaml_file_path = None
    else:
        snapcraft_yaml_file_path = get_snapcraft_yaml()

    ctx = click.get_current_context()
    for key, value in ctx.parent.params.items():
        if not kwargs.get(key):
            kwargs[key] = value

    project = Project(
        debug=kwargs.pop("debug"),
        target_deb_arch=kwargs.pop("target_arch"),
        snapcraft_yaml_file_path=snapcraft_yaml_file_path,
        is_managed_host=is_managed_host,
    )
    return project



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft import formatting_utils
from snapcraft.internal import errors
from typing import List


class CommandError(errors.SnapcraftError):
    def __init__(self, message: str) -> None:
        self.fmt = message


class SnapMetaGenerationError(errors.SnapcraftError):
    pass


class MissingSnapcraftYamlKeysError(SnapMetaGenerationError):

    fmt = (
        "Failed to generate snap metadata: "
        "Missing required key(s) in snapcraft.yaml: {keys}. "
        "Either specify the missing key(s), or use 'adopt-info' to get them "
        "from a part."
    )

    def __init__(self, keys: list) -> None:
        super().__init__(keys=formatting_utils.humanize_list(keys, "and"))


class AdoptedPartMissingError(SnapMetaGenerationError):

    fmt = (
        "Failed to generate snap metadata: "
        "'adopt-info' refers to a part named {part!r}, but it is not defined "
        "in the 'snapcraft.yaml' file."
    )

    def __init__(self, part: str) -> None:
        super().__init__(part=part)


class AdoptedPartNotParsingInfo(SnapMetaGenerationError):

    fmt = (
        "Failed to generate snap metadata: "
        "'adopt-info' refers to part {part!r}, but that part is lacking the "
        "'parse-info' property."
    )

    def __init__(self, part: str) -> None:
        super().__init__(part=part)


class AmbiguousPassthroughKeyError(SnapMetaGenerationError):

    fmt = (
        "Failed to generate snap metadata: "
        "The following keys are specified in their regular location "
        "as well as in passthrough: {keys}. "
        "Remove duplicate keys."
    )

    def __init__(self, keys: List[str]) -> None:
        super().__init__(keys=formatting_utils.humanize_list(keys, "and"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Common 'source' options.

Unless the part plugin overrides this behaviour, a part can use these
'source' keys in its definition. They tell snapcraft where to pull source
code for that part, and how to unpack it if necessary.

  - source: url-or-path

    A URL or path to some source tree to build. It can be local
    ('./src/foo') or remote ('https://foo.org/...'), and can refer to a
    directory tree or a tarball or a revision control repository
    ('git:...').

  - source-type: git, bzr, hg, svn, tar, deb, rpm, or zip

    In some cases the source string is not enough to identify the version
    control system or compression algorithm. The source-type key can tell
    snapcraft exactly how to treat that content.

  - source-checksum: <algorithm>/<digest>

    Snapcraft will use the digest specified to verify the integrity of the
    source. The source-type needs to be a file (tar, zip, deb or rpm) and
    the algorithm either md5, sha1, sha224, sha256, sha384, sha512, sha3_256,
    sha3_384 or sha3_512.

  - source-depth: <integer>

    By default clones or branches with full history, specifying a depth
    will truncate the history to the specified number of commits.

  - source-branch: <branch-name>

    Snapcraft will checkout a specific branch from the source tree. This
    only works on multi-branch repositories from git and hg (mercurial).

  - source-commit: <commit>

    Snapcraft will checkout the specific commit from the source tree revision
    control system.

  - source-tag: <tag>

    Snapcraft will checkout the specific tag from the source tree revision
    control system.

  - source-subdir: path

    When building, Snapcraft will set the working directory to be this
    subdirectory within the source.

Note that plugins might well define their own semantics for the 'source'
keywords, because they handle specific build systems, and many languages
have their own built-in packaging systems (think CPAN, PyPI, NPM). In those
cases you want to refer to the help text for the specific plugin.

  snapcraft help <plugin>

"""
import logging
import os
import os.path
import re
import sys

from . import errors

if sys.platform == "linux":
    from ._bazaar import Bazaar  # noqa
    from ._git import Git  # noqa
    from ._local import Local  # noqa
    from ._mercurial import Mercurial  # noqa
    from ._script import Script  # noqa
    from ._subversion import Subversion  # noqa
    from ._tar import Tar  # noqa
    from ._zip import Zip  # noqa
    from ._7z import SevenZip  # noqa
    from ._deb import Deb  # noqa
    from ._rpm import Rpm  # noqa

    _source_handler = {
        "bzr": Bazaar,
        "git": Git,
        "hg": Mercurial,
        "mercurial": Mercurial,
        "subversion": Subversion,
        "svn": Subversion,
        "tar": Tar,
        "zip": Zip,
        "7z": SevenZip,
        "local": Local,
        "deb": Deb,
        "rpm": Rpm,
        "": Local,
    }


logging.getLogger("urllib3").setLevel(logging.CRITICAL)


__SOURCE_DEFAULTS = {
    "source": None,
    "source-commit": None,
    "source-checksum": None,
    "source-depth": None,
    "source-tag": None,
    "source-type": None,
    "source-branch": None,
    "source-subdir": None,
}


def get_source_defaults():
    return __SOURCE_DEFAULTS.copy()


def get(sourcedir, builddir, options):
    """Populate sourcedir and builddir from parameters defined in options.

    :param str sourcedir: The source directory to use.
    :param str builddir: The build directory to use.
    :param options: source options.
    """
    source_type = getattr(options, "source_type", None)
    source_attributes = dict(
        source_depth=getattr(options, "source_depth", None),
        source_checksum=getattr(options, "source_checksum", None),
        source_tag=getattr(options, "source_tag", None),
        source_commit=getattr(options, "source_commit", None),
        source_branch=getattr(options, "source_branch", None),
    )

    handler_class = get_source_handler(options.source, source_type=source_type)
    handler = handler_class(options.source, sourcedir, **source_attributes)
    handler.pull()


def get_source_handler_from_type(source_type):
    """Return the source handler for source_type."""
    return _source_handler.get(source_type)


def get_source_handler(source, *, source_type=""):
    if not source_type:
        source_type = _get_source_type_from_uri(source)

    return _source_handler[source_type]


_tar_type_regex = re.compile(r".*\.((tar(\.(xz|gz|bz2))?)|tgz)$")


def _get_source_type_from_uri(source, ignore_errors=False):  # noqa: C901
    for extension in ["zip", "deb", "rpm", "7z"]:
        if source.endswith(".{}".format(extension)):
            return extension
    source_type = ""
    if source.startswith("bzr:") or source.startswith("lp:"):
        source_type = "bzr"
    elif (
        source.startswith("git:")
        or source.startswith("git@")
        or source.endswith(".git")
    ):
        source_type = "git"
    elif source.startswith("svn:"):
        source_type = "subversion"
    elif _tar_type_regex.match(source):
        source_type = "tar"
    elif os.path.isdir(source):
        source_type = "local"
    elif not ignore_errors:
        raise errors.SnapcraftSourceUnhandledError(source)

    return source_type



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Set, TYPE_CHECKING

import snapcraft

from .errors import OnStatementSyntaxError
from . import typing
from ._statement import Statement

# Don't use circular imports unless type checking
if TYPE_CHECKING:
    from ._processor import GrammarProcessor  # noqa: F401

_SELECTOR_PATTERN = re.compile(r"\Aon\s+([^,\s](?:,?[^,]+)*)\Z")
_WHITESPACE_PATTERN = re.compile(r"\A.*\s.*\Z")


class OnStatement(Statement):
    """Process an 'on' statement in the grammar.

    For example:
    >>> from snapcraft import ProjectOptions
    >>> from snapcraft.internal.project_loader import grammar
    >>> from unittest import mock
    >>>
    >>> def checker(primitive):
    ...     return True
    >>> options = ProjectOptions()
    >>> processor = grammar.GrammarProcessor(None, options, checker)
    >>>
    >>> clause = OnStatement(on='on amd64', body=['foo'], processor=processor)
    >>> clause.add_else(['bar'])
    >>> with mock.patch('platform.machine') as mock_machine:
    ...     # Pretend this machine is an i686, not amd64
    ...     mock_machine.return_value = 'i686'
    ...     clause.process()
    {'bar'}
    """

    def __init__(
        self,
        *,
        on: str,
        body: typing.Grammar,
        processor: "GrammarProcessor",
        call_stack: typing.CallStack = None
    ) -> None:
        """Create an OnStatement instance.

        :param str on: The 'on <selectors>' part of the clause.
        :param list body: The body of the clause.
        :param GrammarProcessor process: GrammarProcessor to use for processing
                                         this statement.
        :param list call_stack: Call stack leading to this statement.
        """
        super().__init__(body=body, processor=processor, call_stack=call_stack)

        self.selectors = _extract_on_clause_selectors(on)

    def _check(self) -> bool:
        """Check if a statement main body should be processed.

        :return: True if main body should be processed, False if elses should
                 be processed.
        :rtype: bool
        """
        # A new ProjectOptions instance defaults to the host architecture
        # whereas self._project_options would yield the target architecture
        host_arch = snapcraft.ProjectOptions().deb_arch

        # The only selector currently supported is the host arch. Since
        # selectors are matched with an AND, not OR, there should only be one
        # selector.
        return (len(self.selectors) == 1) and (host_arch in self.selectors)

    def __eq__(self, other) -> bool:
        if type(other) is type(self):
            return self.selectors == other.selectors

        return False

    def __str__(self) -> str:
        return "on {}".format(",".join(sorted(self.selectors)))


def _extract_on_clause_selectors(on: str) -> Set[str]:
    """Extract the list of selectors within an on clause.

    :param str on: The 'on <selector>' part of the 'on' clause.

    :return: Selectors found within the 'on' clause.
    :rtype: set

    For example:
    >>> _extract_on_clause_selectors('on amd64,i386') == {'amd64', 'i386'}
    True
    """

    match = _SELECTOR_PATTERN.match(on)

    try:
        selector_group = match.group(1)
    except AttributeError:
        raise OnStatementSyntaxError(on, message="selectors are missing")
    except IndexError:
        raise OnStatementSyntaxError(on)

    # This could be part of the _SELECTOR_PATTERN, but that would require us
    # to provide a very generic error when we can try to be more helpful.
    if _WHITESPACE_PATTERN.match(selector_group):
        raise OnStatementSyntaxError(
            on, message="spaces are not allowed in the selectors"
        )

    return {selector.strip() for selector in selector_group.split(",")}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.errors import MetadataExtractionError


class UnhandledFileError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "This file is not handled by {extractor_name!r}."
    )

    def __init__(self, path: str, extractor_name: str) -> None:
        super().__init__(path=path, extractor_name=extractor_name)


class AppstreamFileParseError(MetadataExtractionError):

    fmt = "Failed to extract metadata from {path!r}: it's not a valid XML file."

    def __init__(self, path: str) -> None:
        super().__init__(path=path)


class SetupPyFileParseError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "the logic in setup.py is currently not handled."
    )

    def __init__(self, path: str) -> None:
        super().__init__(path=path)


class SetupPyImportError(MetadataExtractionError):

    fmt = (
        "Failed to extract metadata from {path!r}: "
        "some packages or modules used could not be imported: "
        "{error}"
    )

    def __init__(self, path: str, error: str) -> None:
        super().__init__(path=path, error=error)
